import pandas as pd

from ..vector_model import VectorModel


class VectorModelWithSeparateFeatureGeneration:
    def __init__(self, vectorModel: VectorModel):
        self.vectorModel = vectorModel
        self.featureGen = vectorModel.getFeatureGenerator()
        self.vectorModel.setFeatureGenerator(None)

    def __str__(self):
        return self.vectorModel.__str__()

    class IntermediateFittingStep:
        def __init__(self, vectorModel: VectorModel, X: pd.DataFrame, Y: pd.DataFrame):
            self.Y = Y
            self.X = X
            self.vectorModel = vectorModel

        def execute(self) -> VectorModel:
            self.vectorModel.fit(self.X, self.Y)
            return self.vectorModel

        def __str__(self):
            return f"{self.__class__.__name__} for {self.vectorModel}"

    class PredictFinaliser:
        def __init__(self, vectorModel: VectorModel, X: pd.DataFrame):
            self.X = X
            self.vectorModel = vectorModel

        def execute(self) -> pd.DataFrame:
            return self.vectorModel.predict(self.X)

        def __str__(self):
            return f"{self.__class__.__name__} for {self.vectorModel}"

    def fitStart(self, X, Y) -> 'VectorModelWithSeparateFeatureGeneration.IntermediateFittingStep':
        X = self.featureGen.fitGenerate(X, Y)
        return self.IntermediateFittingStep(self.vectorModel, X, Y)

    def predictStart(self, X: pd.DataFrame):
        X = self.featureGen.generate(X)
        return self.PredictFinaliser(self.vectorModel, X)

    def fitEnd(self, vectorModel) -> VectorModel:
        vectorModel._featureGenerator = self.featureGen
        return vectorModel



from abc import ABC, abstractmethod
from typing import Tuple, Sequence, TypeVar, List, Generic

import numpy as np
import pandas as pd
import scipy.stats


T = TypeVar("T")


class BaseInputOutputData(Generic[T], ABC):
    def __init__(self, inputs: T, outputs: T):
        """
        :param inputs: expected to have shape and __len__
        :param outputs: expected to have shape and __len__
        """
        if len(inputs) != len(outputs):
            raise ValueError("Lengths do not match")
        self.inputs = inputs
        self.outputs = outputs

    def __len__(self):
        return len(self.inputs)

    @abstractmethod
    def filterIndices(self, indices: Sequence[int]) -> __qualname__:
        pass


class InputOutputArrays(BaseInputOutputData[np.ndarray]):
    def __init__(self, inputs: np.ndarray, outputs: np.ndarray):
        super().__init__(inputs, outputs)

    def filterIndices(self, indices: Sequence[int]) -> __qualname__:
        inputs = self.inputs[indices]
        outputs = self.outputs[indices]
        return InputOutputArrays(inputs, outputs)

    def toTorchDataLoader(self, batchSize=64, shuffle=True):
        try:
            import torch
            from torch.utils.data import DataLoader, TensorDataset
        except ImportError:
            raise ImportError(f"Could not import torch, did you install it?")
        dataSet = TensorDataset(torch.tensor(self.inputs), torch.tensor(self.outputs))
        return DataLoader(dataSet, batch_size=batchSize, shuffle=shuffle)


# TODO: Rename to InputOutputDataFrames when the time for breaking changes has come
class InputOutputData(BaseInputOutputData[pd.DataFrame]):
    def __init__(self, inputs: pd.DataFrame, outputs: pd.DataFrame):
        super().__init__(inputs, outputs)

    def filterIndices(self, indices: Sequence[int]) -> __qualname__:
        inputs = self.inputs.iloc[indices]
        outputs = self.outputs.iloc[indices]
        return InputOutputData(inputs, outputs)

    @property
    def inputDim(self):
        return self.inputs.shape[1]

    @property
    def outputDim(self):
        return self.outputs.shape[1]

    def computeInputOutputCorrelation(self):
        correlations = {}
        for outputCol in self.outputs.columns:
            correlations[outputCol] = {}
            outputSeries = self.outputs[outputCol]
            for inputCol in self.inputs.columns:
                inputSeries = self.inputs[inputCol]
                pcc, pvalue = scipy.stats.pearsonr(inputSeries, outputSeries)
                correlations[outputCol][inputCol] = pcc
        return correlations


TInputOutputData = TypeVar("TInputOutputData", bound=BaseInputOutputData)


class DataSplitter(ABC, Generic[TInputOutputData]):
    @abstractmethod
    def split(self, data: TInputOutputData) -> Tuple[TInputOutputData, TInputOutputData]:
        pass


class DataSplitterFractional(DataSplitter):
    def __init__(self, fractionalSizeOfFirstSet: float, shuffle=True, randomSeed=42):
        if not 0 <= fractionalSizeOfFirstSet <= 1:
            raise Exception(f"invalid fraction: {fractionalSizeOfFirstSet}")
        self.fractionalSizeOfFirstSet = fractionalSizeOfFirstSet
        self.shuffle = shuffle
        self.randomSeed = randomSeed

    def split(self, data: TInputOutputData) -> Tuple[TInputOutputData, TInputOutputData]:
        numDataPoints = len(data)
        splitIndex = int(numDataPoints * self.fractionalSizeOfFirstSet)
        if self.shuffle:
            rand = np.random.RandomState(self.randomSeed)
            indices = rand.permutation(numDataPoints)
        else:
            indices = range(numDataPoints)
        indicesA = indices[:splitIndex]
        indicesB = indices[splitIndex:]
        A = data.filterIndices(list(indicesA))
        B = data.filterIndices(list(indicesB))
        return A, B


class DataSplitterFromDataFrameSplitter(DataSplitter[InputOutputData]):
    """
    Creates a DataSplitter from a DataFrameSplitter, which can be applied either to the input or the output data.
    It supports only InputOutputData, not other subclasses of BaseInputOutputData.
    """
    def __init__(self, dataFrameSplitter: "DataFrameSplitter", fractionalSizeOfFirstSet: float, applyToInput=True):
        """
        :param dataFrameSplitter: the splitter to apply
        :param fractionalSizeOfFirstSet: the desired fractional size of the first set when applying the splitter
        :param applyToInput: if True, apply the splitter to the input data frame; if False, apply it to the output data frame
        """
        self.dataFrameSplitter = dataFrameSplitter
        self.fractionalSizeOfFirstSet = fractionalSizeOfFirstSet
        self.applyToInput = applyToInput

    def split(self, data: InputOutputData) -> Tuple[InputOutputData, InputOutputData]:
        if not isinstance(data, InputOutputData):
            raise ValueError(f"{self} is only applicable to instances of {InputOutputData.__name__}, got {data}")
        df = data.inputs if self.applyToInput else data.outputs
        indicesA, indicesB = self.dataFrameSplitter.computeSplitIndices(df, self.fractionalSizeOfFirstSet)
        A = data.filterIndices(list(indicesA))
        B = data.filterIndices(list(indicesB))
        return A, B


class DataFrameSplitter(ABC):
    @abstractmethod
    def computeSplitIndices(self, df: pd.DataFrame, fractionalSizeOfFirstSet: float) -> Tuple[Sequence[int], Sequence[int]]:
        pass

    @staticmethod
    def split(df: pd.DataFrame, indicesPair: Tuple[Sequence[int], Sequence[int]]) -> Tuple[pd.DataFrame, pd.DataFrame]:
        indicesA, indicesB = indicesPair
        A = df.iloc[indicesA]
        B = df.iloc[indicesB]
        return A, B


class DataFrameSplitterFractional(DataFrameSplitter):
    def __init__(self, shuffle=False, randomSeed=42):
        self.randomSeed = randomSeed
        self.shuffle = shuffle

    def computeSplitIndices(self, df: pd.DataFrame, fractionalSizeOfFirstSet: float) -> Tuple[Sequence[int], Sequence[int]]:
        n = df.shape[0]
        sizeA = int(n * fractionalSizeOfFirstSet)
        if self.shuffle:
            rand = np.random.RandomState(self.randomSeed)
            indices = rand.permutation(n)
        else:
            indices = list(range(n))
        indices_A = indices[:sizeA]
        indices_B = indices[sizeA:]
        return indices_A, indices_B



import functools
import logging
from abc import ABC, abstractmethod
from typing import Tuple, Dict, Any, Generator, Generic, TypeVar, Sequence, Optional, List, Union, Callable

import pandas as pd

from .eval_stats.eval_stats_base import EvalStats, EvalStatsCollection
from .eval_stats.eval_stats_classification import ClassificationEvalStats, ClassificationMetric
from .eval_stats.eval_stats_regression import RegressionEvalStats, RegressionEvalStatsCollection, RegressionMetric
from ..data_transformation import DataFrameTransformer
from ..data import DataSplitter, DataSplitterFractional, InputOutputData
from ..tracking import TrackingMixin
from ..util.string import ToStringMixin
from ..util.typing import PandasNamedTuple
from ..vector_model import VectorClassificationModel, VectorModel, VectorModelBase, VectorModelFittableBase

log = logging.getLogger(__name__)

TModel = TypeVar("TModel", bound=VectorModel)
TEvalStats = TypeVar("TEvalStats", bound=EvalStats)
TEvalStatsCollection = TypeVar("TEvalStatsCollection", bound=EvalStatsCollection)


class MetricsDictProvider(TrackingMixin, ABC):
    @abstractmethod
    def _computeMetrics(self, model, **kwargs) -> Dict[str, float]:
        """
        Computes metrics for the given model, typically by fitting the model and applying it to test data

        :param model: the model
        :param kwargs: parameters to pass on to the underlying evaluation method
        :return: a dictionary with metrics values
        """
        pass

    def computeMetrics(self, model, **kwargs) -> Optional[Dict[str, float]]:
        """
        Computes metrics for the given model, typically by fitting the model and applying it to test data.
        If a tracked experiment was previously set, the metrics are tracked with the string representation
        of the model added under an additional key 'str(model)'.

        :param model: the model for which to compute metrics
        :param kwargs: parameters to pass on to the underlying evaluation method
        :return: a dictionary with metrics values
        """
        valuesDict = self._computeMetrics(model, **kwargs)
        if self.trackedExperiment is not None:
            trackedDict = valuesDict.copy()
            trackedDict["str(model)"] = str(model)
            self.trackedExperiment.trackValues(trackedDict)
        return valuesDict


class MetricsDictProviderFromFunction(MetricsDictProvider):
    def __init__(self, computeMetricsFn: Callable[[VectorModel], Dict[str, float]]):
        self._computeMetricsFn = computeMetricsFn

    def _computeMetrics(self, model, **kwargs) -> Dict[str, float]:
        return self._computeMetricsFn(model)


class VectorModelEvaluationData(ABC, Generic[TEvalStats]):
    def __init__(self, statsDict: Dict[str, TEvalStats], inputData: pd.DataFrame, model: VectorModelBase):
        """
        :param statsDict: a dictionary mapping from output variable name to the evaluation statistics object
        :param inputData: the input data that was used to produce the results
        :param model: the model that was used to produce predictions
        """
        self.inputData = inputData
        self.evalStatsByVarName = statsDict
        self.predictedVarNames = list(self.evalStatsByVarName.keys())
        self.model = model

    @property
    def modelName(self):
        return self.model.getName()

    def getEvalStats(self, predictedVarName=None) -> TEvalStats:
        if predictedVarName is None:
            if len(self.evalStatsByVarName) != 1:
                raise Exception(f"Must provide name of predicted variable name, as multiple variables were predicted {list(self.evalStatsByVarName.keys())}")
            else:
                predictedVarName = next(iter(self.evalStatsByVarName.keys()))
        evalStats = self.evalStatsByVarName.get(predictedVarName)
        if evalStats is None:
            raise ValueError(f"No evaluation data present for '{predictedVarName}'; known output variables: {list(self.evalStatsByVarName.keys())}")
        return evalStats

    def getDataFrame(self):
        """
        Returns an DataFrame with all evaluation metrics (one row per output variable)

        :return: a DataFrame containing evaluation metrics
        """
        statsDicts = []
        varNames = []
        for predictedVarName, evalStats in self.evalStatsByVarName.items():
            statsDicts.append(evalStats.getAll())
            varNames.append(predictedVarName)
        df = pd.DataFrame(statsDicts, index=varNames)
        df.index.name = "predictedVar"
        return df

    def iterInputOutputGroundTruthTuples(self, predictedVarName=None) -> Generator[Tuple[PandasNamedTuple, Any, Any], None, None]:
        evalStats = self.getEvalStats(predictedVarName)
        for i, namedTuple in enumerate(self.inputData.itertuples()):
            yield namedTuple, evalStats.y_predicted[i], evalStats.y_true[i]


class VectorRegressionModelEvaluationData(VectorModelEvaluationData[RegressionEvalStats]):
    def getEvalStatsCollection(self):
        return RegressionEvalStatsCollection(list(self.evalStatsByVarName.values()))


TEvalData = TypeVar("TEvalData", bound=VectorModelEvaluationData)


class VectorModelEvaluatorParams(ToStringMixin, ABC):
    def __init__(self, dataSplitter: DataSplitter = None, fractionalSplitTestFraction: float = None, fractionalSplitRandomSeed=42,
            fractionalSplitShuffle=True):
        """
        :param dataSplitter: [if test data must be obtained via split] a splitter to use in order to obtain; if None, must specify
            fractionalSplitTestFraction for fractional split (default)
        :param fractionalSplitTestFraction: [if test data must be obtained via split, dataSplitter is None] the fraction of the data to use for testing/evaluation;
        :param fractionalSplitRandomSeed: [if test data must be obtained via split, dataSplitter is none] the random seed to use for the fractional split of the data
        :param fractionalSplitShuffle: [if test data must be obtained via split, dataSplitter is None] whether to randomly (based on randomSeed) shuffle the dataset before
            splitting it
        """
        self._dataSplitter = dataSplitter
        self._fractionalSplitTestFraction = fractionalSplitTestFraction
        self._fractionalSplitRandomSeed = fractionalSplitRandomSeed
        self._fractionalSplitShuffle = fractionalSplitShuffle

    def _toStringExcludePrivate(self) -> bool:
        return True

    def _toStringAdditionalEntries(self) -> Dict[str, Any]:
        d = {}
        if self._dataSplitter is not None:
            d["dataSplitter"] = self._dataSplitter
        else:
            d["fractionalSplitTestFraction"] = self._fractionalSplitTestFraction
            d["fractionalSplitRandomSeed"] = self._fractionalSplitRandomSeed
            d["fractionalSplitShuffle"] = self._fractionalSplitShuffle
        return d

    def getDataSplitter(self) -> DataSplitter:
        if self._dataSplitter is None:
            if self._fractionalSplitTestFraction is None:
                raise ValueError("Cannot create default data splitter, as no split fraction was provided")
            self._dataSplitter = DataSplitterFractional(1 - self._fractionalSplitTestFraction, shuffle=self._fractionalSplitShuffle,
                randomSeed=self._fractionalSplitRandomSeed)
        return self._dataSplitter


class VectorModelEvaluator(MetricsDictProvider, Generic[TEvalData], ABC):
    def __init__(self, data: Optional[InputOutputData], testData: InputOutputData = None, params: VectorModelEvaluatorParams = None):
        """
        Constructs an evaluator with test and training data.

        :param data: the full data set, or, if testData is given, the training data
        :param testData: the data to use for testing/evaluation; if None, must specify appropriate parameters to define splitting
        :param params: the parameters
        """
        if testData is None:
            if params is None:
                raise ValueError("Parameters required for data split must be provided")
            dataSplitter = params.getDataSplitter()
            self.trainingData, self.testData = dataSplitter.split(data)
            log.debug(f"{dataSplitter} created split with {len(self.trainingData)} ({100*len(self.trainingData)/len(data):.2f}%) and "
                f"{len(self.testData)} ({100*len(self.testData)/len(data):.2f}%) training and test data points respectively")
        else:
            self.trainingData = data
            self.testData = testData

    def evalModel(self, model: VectorModelBase, onTrainingData=False) -> TEvalData:
        """
        Evaluates the given model

        :param model: the model to evaluate
        :param onTrainingData: if True, evaluate on this evaluator's training data rather than the held-out test data
        :return: the evaluation result
        """
        data = self.trainingData if onTrainingData else self.testData
        return self._evalModel(model, data)

    @abstractmethod
    def _evalModel(self, model: VectorModelBase, data: InputOutputData) -> TEvalData:
        pass

    def _computeMetrics(self, model: VectorModel, onTrainingData=False) -> Dict[str, float]:
        return self._computeMetricsForVarName(model, None, onTrainingData=onTrainingData)

    def _computeMetricsForVarName(self, model, predictedVarName: Optional[str], onTrainingData=False):
        self.fitModel(model)
        evalData = self.evalModel(model, onTrainingData=onTrainingData)
        return evalData.getEvalStats(predictedVarName=predictedVarName).getAll()

    def createMetricsDictProvider(self, predictedVarName: Optional[str]) -> MetricsDictProvider:
        """
        Creates a metrics dictionary provider, e.g. for use in hyperparameter optimisation

        :param predictedVarName: the name of the predicted variable for which to obtain evaluation metrics; may be None only
            if the model outputs but a single predicted variable
        :return: a metrics dictionary provider instance for the given variable
        """
        return MetricsDictProviderFromFunction(functools.partial(self._computeMetricsForVarName, predictedVarName=predictedVarName))

    def fitModel(self, model: VectorModelFittableBase):
        """Fits the given model's parameters using this evaluator's training data"""
        if self.trainingData is None:
            raise Exception(f"Cannot fit model with evaluator {self.__class__.__name__}: no training data provided")
        model.fit(self.trainingData.inputs, self.trainingData.outputs)


class VectorRegressionModelEvaluatorParams(VectorModelEvaluatorParams):
    def __init__(self, dataSplitter: DataSplitter = None, fractionalSplitTestFraction: float = None, fractionalSplitRandomSeed=42,
            fractionalSplitShuffle=True, additionalMetrics: Sequence[RegressionMetric] = None,
            outputDataFrameTransformer: DataFrameTransformer = None):
        """
        :param dataSplitter: [if test data must be obtained via split] a splitter to use in order to obtain; if None, must specify
            fractionalSplitTestFraction for fractional split (default)
        :param fractionalSplitTestFraction: [if dataSplitter is None, test data must be obtained via split] the fraction of the data to use for testing/evaluation;
        :param fractionalSplitRandomSeed: [if dataSplitter is none, test data must be obtained via split] the random seed to use for the fractional split of the data
        :param fractionalSplitShuffle: [if dataSplitter is None, test data must be obtained via split] whether to randomly (based on randomSeed) shuffle the dataset before
            splitting it
        :param additionalMetrics: additional regression metrics to apply
        :param outputDataFrameTransformer: a data frame transformer to apply to all output data frames (both model outputs and ground truth),
            such that evaluation metrics are computed on the transformed data frame
        """
        super().__init__(dataSplitter, fractionalSplitTestFraction=fractionalSplitTestFraction, fractionalSplitRandomSeed=fractionalSplitRandomSeed,
            fractionalSplitShuffle=fractionalSplitShuffle)
        self.additionalMetrics = additionalMetrics
        self.outputDataFrameTransformer = outputDataFrameTransformer

    @classmethod
    def fromDictOrInstance(cls, params: Optional[Union[Dict[str, Any], "VectorRegressionModelEvaluatorParams"]]) -> "VectorRegressionModelEvaluatorParams":
        if params is None:
            return VectorRegressionModelEvaluatorParams()
        elif type(params) == dict:
            return cls.fromOldKwArgs(**params)
        elif isinstance(params, VectorRegressionModelEvaluatorParams):
            return params
        else:
            raise ValueError(f"Must provide dictionary or instance, got {params}")

    @classmethod
    def fromOldKwArgs(cls, dataSplitter=None, testFraction=None, randomSeed=42, shuffle=True, additionalMetrics: Sequence[RegressionMetric] = None,
            outputDataFrameTransformer: DataFrameTransformer = None) -> "VectorRegressionModelEvaluatorParams":
        return cls(dataSplitter=dataSplitter, fractionalSplitTestFraction=testFraction, fractionalSplitRandomSeed=randomSeed,
            fractionalSplitShuffle=shuffle, additionalMetrics=additionalMetrics, outputDataFrameTransformer=outputDataFrameTransformer)


class VectorRegressionModelEvaluator(VectorModelEvaluator[VectorRegressionModelEvaluationData]):
    def __init__(self, data: Optional[InputOutputData], testData: InputOutputData = None, params: VectorRegressionModelEvaluatorParams = None,
            **kwArgsOldParams):
        """
        Constructs an evaluator with test and training data.

        :param data: the full data set, or, if testData is given, the training data
        :param testData: the data to use for testing/evaluation; if None, must specify appropriate parameters to define splitting
        :param params: the parameters
        :param kwArgsOldParams: old-style keyword parameters (for backward compatibility only)
        """
        params = self._createParams(params, kwArgsOldParams)
        super().__init__(data=data, testData=testData, params=params)
        self.params = params

    @staticmethod
    def _createParams(params: VectorRegressionModelEvaluatorParams, kwArgsOldParams: dict) -> VectorRegressionModelEvaluatorParams:
        if params is not None:
            return params
        elif len(kwArgsOldParams) > 0:
            return VectorRegressionModelEvaluatorParams.fromOldKwArgs(**kwArgsOldParams)
        else:
            return VectorRegressionModelEvaluatorParams()

    def _evalModel(self, model: VectorModelBase, data: InputOutputData) -> VectorRegressionModelEvaluationData:
        if not model.isRegressionModel():
            raise ValueError(f"Expected a regression model, got {model}")
        evalStatsByVarName = {}
        predictions, groundTruth = self._computeOutputs(model, data)
        for predictedVarName in predictions.columns:
            evalStats = RegressionEvalStats(y_predicted=predictions[predictedVarName], y_true=groundTruth[predictedVarName],
                additionalMetrics=self.params.additionalMetrics)
            evalStatsByVarName[predictedVarName] = evalStats
        return VectorRegressionModelEvaluationData(evalStatsByVarName, data.inputs, model)

    def computeTestDataOutputs(self, model: VectorModelBase) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """
        Applies the given model to the test data

        :param model: the model to apply
        :return: a pair (predictions, groundTruth)
        """
        return self._computeOutputs(model, self.testData)

    def _computeOutputs(self, model: VectorModelBase, inputOutputData: InputOutputData):
        """
        Applies the given model to the given data

        :param model: the model to apply
        :param inputOutputData: the data set
        :return: a pair (predictions, groundTruth)
        """
        predictions = model.predict(inputOutputData.inputs)
        groundTruth = inputOutputData.outputs
        if self.params.outputDataFrameTransformer:
            predictions = self.params.outputDataFrameTransformer.apply(predictions)
            groundTruth = self.params.outputDataFrameTransformer.apply(groundTruth)
        return predictions, groundTruth


class VectorClassificationModelEvaluationData(VectorModelEvaluationData[ClassificationEvalStats]):
    pass


class VectorClassificationModelEvaluatorParams(VectorModelEvaluatorParams):
    def __init__(self, dataSplitter: DataSplitter = None, fractionalSplitTestFraction: float = None, fractionalSplitRandomSeed=42,
            fractionalSplitShuffle=True, additionalMetrics: Sequence[ClassificationMetric] = None,
            computeProbabilities: bool = False):
        """
        :param dataSplitter: [if test data must be obtained via split] a splitter to use in order to obtain; if None, must specify
            fractionalSplitTestFraction for fractional split (default)
        :param fractionalSplitTestFraction: [if dataSplitter is None, test data must be obtained via split] the fraction of the data to use for testing/evaluation;
        :param fractionalSplitRandomSeed: [if dataSplitter is none, test data must be obtained via split] the random seed to use for the fractional split of the data
        :param fractionalSplitShuffle: [if dataSplitter is None, test data must be obtained via split] whether to randomly (based on randomSeed) shuffle the dataset before
            splitting it
        :param additionalMetrics: additional metrics to apply
        :param computeProbabilities: whether to compute class probabilities
        """
        super().__init__(dataSplitter, fractionalSplitTestFraction=fractionalSplitTestFraction, fractionalSplitRandomSeed=fractionalSplitRandomSeed,
            fractionalSplitShuffle=fractionalSplitShuffle)
        self.additionalMetrics = additionalMetrics
        self.computeProbabilities = computeProbabilities

    @classmethod
    def fromOldKwArgs(cls, dataSplitter=None, testFraction=None,
            randomSeed=42, computeProbabilities=False, shuffle=True, additionalMetrics: Sequence[ClassificationMetric] = None):
        return cls(dataSplitter=dataSplitter, fractionalSplitTestFraction=testFraction, fractionalSplitRandomSeed=randomSeed,
            fractionalSplitShuffle=shuffle, additionalMetrics=additionalMetrics, computeProbabilities=computeProbabilities)

    @classmethod
    def fromDictOrInstance(cls, params: Optional[Union[Dict[str, Any], "VectorClassificationModelEvaluatorParams"]]) -> "VectorClassificationModelEvaluatorParams":
        if params is None:
            return VectorClassificationModelEvaluatorParams()
        elif type(params) == dict:
            return cls.fromOldKwArgs(**params)
        elif isinstance(params, VectorClassificationModelEvaluatorParams):
            return params
        else:
            raise ValueError(f"Must provide dictionary or instance, got {params}")


class VectorClassificationModelEvaluator(VectorModelEvaluator[VectorClassificationModelEvaluationData]):
    def __init__(self, data: Optional[InputOutputData], testData: InputOutputData = None, params: VectorClassificationModelEvaluatorParams = None,
            **kwArgsOldParams):
        """
        Constructs an evaluator with test and training data.

        :param data: the full data set, or, if testData is given, the training data
        :param testData: the data to use for testing/evaluation; if None, must specify appropriate parameters to define splitting
        :param params: the parameters
        :param kwArgsOldParams: old-style keyword parameters (for backward compatibility only)
        """
        params = self._createParams(params, kwArgsOldParams)
        super().__init__(data=data, testData=testData, params=params)
        self.params = params

    @staticmethod
    def _createParams(params: VectorClassificationModelEvaluatorParams, kwArgs: dict) -> VectorClassificationModelEvaluatorParams:
        if params is not None:
            return params
        elif len(kwArgs) > 0:
            return VectorClassificationModelEvaluatorParams.fromOldKwArgs(**kwArgs)
        else:
            return VectorClassificationModelEvaluatorParams()

    def _evalModel(self, model: VectorClassificationModel, data: InputOutputData) -> VectorClassificationModelEvaluationData:
        if model.isRegressionModel():
            raise ValueError(f"Expected a classification model, got {model}")
        predictions, predictions_proba, groundTruth = self._computeOutputs(model, data)
        evalStats = ClassificationEvalStats(y_predictedClassProbabilities=predictions_proba, y_predicted=predictions, y_true=groundTruth,
            labels=model.getClassLabels(), additionalMetrics=self.params.additionalMetrics)
        predictedVarName = model.getPredictedVariableNames()[0]
        return VectorClassificationModelEvaluationData({predictedVarName: evalStats}, data.inputs, model)

    def computeTestDataOutputs(self, model) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        """
        Applies the given model to the test data

        :param model: the model to apply
        :return: a triple (predictions, predicted class probability vectors, groundTruth) of DataFrames
        """
        return self._computeOutputs(model, self.testData)

    def _computeOutputs(self, model, inputOutputData: InputOutputData) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        """
        Applies the given model to the given data

        :param model: the model to apply
        :param inputOutputData: the data set
        :return: a triple (predictions, predicted class probability vectors, groundTruth) of DataFrames
        """
        if self.params.computeProbabilities:
            classProbabilities = model.predictClassProbabilities(inputOutputData.inputs)
            predictions = model.convertClassProbabilitiesToPredictions(classProbabilities)
        else:
            classProbabilities = None
            predictions = model.predict(inputOutputData.inputs)
        groundTruth = inputOutputData.outputs
        return predictions, classProbabilities, groundTruth


class RuleBasedVectorClassificationModelEvaluator(VectorClassificationModelEvaluator):
    def __init__(self, data: InputOutputData):
        super().__init__(data, testData=data)

    def evalModel(self, model: VectorModelBase, onTrainingData=False) -> VectorClassificationModelEvaluationData:
        """
        Evaluate the rule based model. The training data and test data coincide, thus fitting the model
        will fit the model's preprocessors on the full data set and evaluating it will evaluate the model on the
        same data set.

        :param model:
        :param onTrainingData: has to be False here. Setting to True is not supported and will lead to an
            exception
        :return:
        """
        if onTrainingData:
            raise Exception("Evaluating rule based models on training data is not supported. In this evaluator"
                            "training and test data coincide.")
        return super().evalModel(model)


class RuleBasedVectorRegressionModelEvaluator(VectorRegressionModelEvaluator):
    def __init__(self, data: InputOutputData):
        super().__init__(data, testData=data)

    def evalModel(self, model: VectorModelBase, onTrainingData=False) -> VectorRegressionModelEvaluationData:
        """
        Evaluate the rule based model. The training data and test data coincide, thus fitting the model
        will fit the model's preprocessors on the full data set and evaluating it will evaluate the model on the
        same data set.

        :param model:
        :param onTrainingData: has to be False here. Setting to True is not supported and will lead to an
            exception
        :return:
        """
        if onTrainingData:
            raise Exception("Evaluating rule based models on training data is not supported. In this evaluator"
                            "training and test data coincide.")
        return super().evalModel(model)



import collections
import logging
import re
from typing import Optional, Union

import pandas as pd
import torch

from .lstnet_modules import LSTNetwork
from ...torch_base import TorchVectorClassificationModel, VectorTorchModel, ClassificationOutputMode
from ...torch_data import TorchDataSetProviderFromDataUtil, TensorScalerIdentity, TensorScaler, DataUtil
from ...torch_enums import ActivationFunction
from ...torch_opt import NNOptimiserParams
from ....util.string import objectRepr

log: logging.Logger = logging.getLogger(__name__)


class LSTNetworkVectorClassificationModel(TorchVectorClassificationModel):
    """
    Classification model for time series data using the LSTNetwork architecture.

    Since the model takes a time series as input, it requires that input data frames to use special naming of columns
    such that the data can be interpreted correctly:
    Each column name must start with an N-digit prefix indicating the time slice the data pertains to (for any fixed N);
    the following suffix shall indicate the name of the actual feature.
    For each N-digit prefix, we must have the same set of suffixes in the list of columns, i.e. we must have the same
    features for each time slice in the input time series.
    """
    def __init__(self, numInputTimeSlices, inputDimPerTimeSlice, numClasses: Optional[int] = None,
            numConvolutions: int = 100, numCnnTimeSlices: int = 6, hidRNN: int = 100, skip: int = 0, hidSkip: int = 5,
            hwWindow: int = 0, hwCombine: str = "plus", dropout=0.2, outputActivation=ActivationFunction.LOG_SOFTMAX, cuda=True,
            nnOptimiserParams: Union[dict, NNOptimiserParams] = None):
        """
        :param numInputTimeSlices: the number of input time slices
        :param inputDimPerTimeSlice: the dimension of the input data per time slice
        :param numClasses: the number of classes considered by this classification problem; if None, determine from data
        :param numCnnTimeSlices: the number of time slices considered by each convolution (i.e. it is one of the dimensions of the matrix used for
            convolutions, the other dimension being inputDimPerTimeSlice), a.k.a. "Ck"
        :param numConvolutions: the number of separate convolutions to apply, i.e. the number of independent convolution matrices, a.k.a "hidC";
            if it is 0, then the entire complex processing path is not applied.
        :param hidRNN: the number of hidden output dimensions for the RNN stage
        :param skip: the number of time slices to skip for the skip-RNN. If it is 0, then the skip-RNN is not used.
        :param hidSkip: the number of output dimensions of each of the skip parallel RNNs
        :param hwWindow: the number of time slices from the end of the input time series to consider as input for the highway component.
            If it is 0, the highway component is not used.
        :param hwCombine: {"plus", "product", "bilinear"} the function with which the highway component's output is combined with the complex path's output
        :param dropout: the dropout probability to use during training (dropouts are applied after every major step in the evaluation path)
        :param outputActivation: the output activation function
        :param nnOptimiserParams: parameters for NNOptimiser to use for training
        """
        self.cuda = cuda
        self.numClasses = numClasses
        lstnetArgs = dict(numInputTimeSlices=numInputTimeSlices, inputDimPerTimeSlice=inputDimPerTimeSlice, numOutputTimeSlices=1,
            outputDimPerTimeSlice=numClasses, numConvolutions=numConvolutions, numCnnTimeSlices=numCnnTimeSlices, hidRNN=hidRNN,
            hwWindow=hwWindow, hwCombine=hwCombine, dropout=dropout, outputActivation=outputActivation,
            skip=skip, hidSkip=hidSkip, isClassification=True)
        outputMode = ClassificationOutputMode.forActivationFn(ActivationFunction.torchFunctionFromAny(outputActivation))
        super().__init__(outputMode, self._LSTNetworkModel, modelArgs=[self.cuda], modelKwArgs=lstnetArgs, nnOptimiserParams=nnOptimiserParams)

    class _LSTNetworkModel(VectorTorchModel):
        def __init__(self, cuda, **lstnetArgs):
            """
            :param cuda: flag indicating whether cuda is used
            :param inputDim: the total number of inputs per data point
            :param numClasses: the number of classes to predict
            :param lstnetArgs: arguments with which to construct the underlying LSTNetwork instance
            """
            super().__init__(cuda)
            self.lstnetArgs = lstnetArgs

        def createTorchModuleForDims(self, inputDim, outputDim):
            expectedInputDim = self.lstnetArgs["numInputTimeSlices"] * self.lstnetArgs["inputDimPerTimeSlice"]
            if expectedInputDim != inputDim:
                raise ValueError(f"Unexpected input size {inputDim}, expected {self.inputDim}")
            if self.lstnetArgs["outputDimPerTimeSlice"] is None:
                self.lstnetArgs["outputDimPerTimeSlice"] = outputDim
            else:
                if self.lstnetArgs["outputDimPerTimeSlice"] != outputDim:
                    raise ValueError(f"Unexpected output size {outputDim}, expected {self.lstnetArgs['outputDimPerTimeSlice']}")
            return LSTNetwork(**self.lstnetArgs)

        def __str__(self):
            return objectRepr(self, self.lstnetArgs)

    def _createDataSetProvider(self, inputs: pd.DataFrame, outputs: pd.DataFrame) -> TorchDataSetProviderFromDataUtil:
        if self.numClasses is None:
            self.numClasses = len(self._labels)
        elif self.numClasses != len(self._labels):
            raise ValueError(f"Output dimension {self.numClasses} per time time slice was specified, while the training data contains {len(self._labels)} classes")
        return TorchDataSetProviderFromDataUtil(self.DataUtil(inputs, outputs, self.numClasses), self.cuda)

    def _predictOutputsForInputDataFrame(self, inputs: pd.DataFrame) -> torch.Tensor:
        log.info(f"Predicting outputs for {len(inputs)} inputs")
        result = super()._predictOutputsForInputDataFrame(inputs)
        return result.squeeze(2)

    def _computeModelInputs(self, x: pd.DataFrame, Y: pd.DataFrame = None, fit=False) -> pd.DataFrame:
        x = super()._computeModelInputs(x, Y=Y, fit=fit)

        # sort input data frame columns by name
        x = x[sorted(x.columns)]

        # check input column name format and consistency
        colNameRegex = re.compile(r"(\d+).+")
        colsByTimeSlice = collections.defaultdict(list)
        numDigits = None
        for colName in x.columns:
            match = colNameRegex.fullmatch(colName)
            if not match:
                raise ValueError(f"Column name '{colName}' does not match the required format (N-digit prefix indicating the time slice order followed by feature name; for any fixed N); columns={list(x.columns)}")
            timeSlice = match.group(1)
            if numDigits is None:
                numDigits = len(timeSlice)
            elif numDigits != len(timeSlice):
                raise ValueError(f"Inconsistent number of digits in column names: Got {numDigits} leading digits for one feature and {len(timeSlice)} for another; columns={list(x.columns)}")
            colsByTimeSlice[timeSlice].append(colName[numDigits:])
        referenceCols = None
        for timeSlice, cols in colsByTimeSlice.items():
            if referenceCols is None:
                referenceCols = cols
            elif referenceCols != cols:
                raise ValueError(f"Inconsistent features across time slices: Got suffixes {cols} for one time slice and {referenceCols} for another; columns={list(x.columns)}")

        return x

    class DataUtil(DataUtil):
        def __init__(self, x_data: pd.DataFrame, y_data: pd.DataFrame, numClasses):
            self.y_data = y_data
            self.x_data = x_data
            self.numClasses = numClasses
            self.scaler = TensorScalerIdentity()

        def inputDim(self):
            return len(self.x_data.columns)

        def modelOutputDim(self) -> int:
            return self.numClasses

        def splitIntoTensors(self, fractionalSizeOfFirstSet):
            splitIndex = round(fractionalSizeOfFirstSet * len(self.y_data))
            y1, x1 = self.getInputOutputPair(self.y_data[:splitIndex], self.x_data[:splitIndex])
            y2, x2 = self.getInputOutputPair(self.y_data[splitIndex:], self.x_data[splitIndex:])
            return (x1, y1), (x2, y2)

        def getInputOutputPair(self, output, input):
            y = torch.tensor(output.values).long()
            x = torch.tensor(input.values).float()
            return y, x

        def getOutputTensorScaler(self) -> TensorScaler:
            return self.scaler

        def getInputTensorScaler(self) -> TensorScaler:
            return self.scaler



import functools
from typing import Union

import pandas as pd

from . import VectorRegressionModel
from .vector_model import RuleBasedVectorRegressionModel


class MultiVectorRegressionModel(RuleBasedVectorRegressionModel):
    """
    Combines several (previously trained) regression models into a single regression model that produces the combined output of the
    individual models (concatenating their outputs)
    """
    def __init__(self, *models: VectorRegressionModel):
        self.models = models
        predictedVariableNamesList = [m.getPredictedVariableNames() for m in models]
        predictedVariableNames = functools.reduce(lambda x, y: x + y.getPredictedVariableNames(), models, [])
        if len(predictedVariableNames) != sum((len(v) for v in predictedVariableNamesList)):
            raise ValueError(f"Models do not produce disjoint outputs: {predictedVariableNamesList}")
        super().__init__(predictedVariableNames)

    def _predict(self, x: pd.DataFrame) -> Union[pd.DataFrame, list]:
        dfs = [m.predict(x) for m in self.models]
        combinedDF = pd.concat(dfs, axis=1)
        return combinedDF


from abc import ABC, abstractmethod
from typing import Dict, Any


class TrackedExperiment(ABC):
    def __init__(self, additionalLoggingValuesDict=None):
        """
        Base class for tracking
        :param additionalLoggingValuesDict: additional values to be logged for each run
        """
        self.additionalLoggingValuesDict = additionalLoggingValuesDict

    def trackValues(self, valuesDict: Dict[str, Any]):
        valuesDict = dict(valuesDict)
        if self.additionalLoggingValuesDict is not None:
            valuesDict.update(self.additionalLoggingValuesDict)
        self._trackValues(valuesDict)

    @abstractmethod
    def _trackValues(self, valuesDict):
        pass


class TrackingMixin(ABC):
    # hackidy hack
    _trackedExperimentAttributeName = "_trackedExperiment"

    def setTrackedExperiment(self, trackedExperiment: TrackedExperiment):
        setattr(self, self._trackedExperimentAttributeName, trackedExperiment)

    def unsetTrackedExperiment(self):
        setattr(self, self._trackedExperimentAttributeName, None)

    @property
    def trackedExperiment(self):
        return getattr(self, self._trackedExperimentAttributeName, None)



import copy
import functools
import logging
from abc import ABC, abstractmethod
from typing import Tuple, Any, Generator, Generic, TypeVar, List, Union, Sequence, Iterable, Optional, Dict

import numpy as np

from .eval_stats.eval_stats_base import PredictionEvalStats, EvalStatsCollection
from .eval_stats.eval_stats_classification import ClassificationEvalStats, ClassificationEvalStatsCollection
from .eval_stats.eval_stats_regression import RegressionEvalStats, RegressionEvalStatsCollection
from .evaluator import VectorRegressionModelEvaluationData, VectorClassificationModelEvaluationData, \
    VectorModelEvaluationData, VectorClassificationModelEvaluator, VectorRegressionModelEvaluator, \
    MetricsDictProvider, VectorModelEvaluator, VectorModelEvaluatorParams, VectorClassificationModelEvaluatorParams, \
    VectorRegressionModelEvaluatorParams, MetricsDictProviderFromFunction
from ..data import InputOutputData
from ..util.typing import PandasNamedTuple
from ..vector_model import VectorClassificationModel, VectorRegressionModel, VectorModel

log = logging.getLogger(__name__)

TModel = TypeVar("TModel", bound=VectorModel)
TEvalStats = TypeVar("TEvalStats", bound=PredictionEvalStats)
TEvalStatsCollection = TypeVar("TEvalStatsCollection", bound=EvalStatsCollection)
TEvalData = TypeVar("TEvalData", bound=VectorModelEvaluationData)


class VectorModelCrossValidationData(ABC, Generic[TModel, TEvalData, TEvalStats, TEvalStatsCollection]):
    def __init__(self, trainedModels: Optional[List[TModel]], evalDataList: List[TEvalData], predictedVarNames: List[str], testIndicesList=None):
        self.predictedVarNames = predictedVarNames
        self.trainedModels = trainedModels
        self.evalDataList = evalDataList
        self.testIndicesList = testIndicesList

    @property
    def modelName(self):
        return self.evalDataList[0].modelName

    @abstractmethod
    def _createEvalStatsCollection(self, l: List[TEvalStats]) -> TEvalStatsCollection:
        pass

    def getEvalStatsCollection(self, predictedVarName=None) -> TEvalStatsCollection:
        if predictedVarName is None:
            if len(self.predictedVarNames) != 1:
                raise Exception(f"Must provide name of predicted variable name, as multiple variables were predicted: {self.predictedVarNames}")
            else:
                predictedVarName = self.predictedVarNames[0]
        evalStatsList = [evalData.getEvalStats(predictedVarName) for evalData in self.evalDataList]
        return self._createEvalStatsCollection(evalStatsList)

    def iterInputOutputGroundTruthTuples(self, predictedVarName=None) -> Generator[Tuple[PandasNamedTuple, Any, Any], None, None]:
        for evalData in self.evalDataList:
            evalStats = evalData.getEvalStats(predictedVarName)
            for i, namedTuple in enumerate(evalData.inputData.itertuples()):
                yield namedTuple, evalStats.y_predicted[i], evalStats.y_true[i]


TCrossValData = TypeVar("TCrossValData", bound=VectorModelCrossValidationData)


class CrossValidationSplitter(ABC):
    """
    Defines a mechanism with which to generate data splits for cross-validation
    """
    @abstractmethod
    def createFolds(self, data: InputOutputData, numFolds: int) -> List[Tuple[Sequence[int], Sequence[int]]]:
        """
        :param data: the data from which to obtain the folds
        :param numFolds: the number of splits/folds
        :return: a list containing numSplits tuples (t, e) where t and e are sequences of data point indices to use for training
            and evaluation respectively
        """
        pass


class CrossValidationSplitterDefault(CrossValidationSplitter):
    def __init__(self, shuffle=True, randomSeed=42):
        self.shuffle = shuffle
        self.randomSeed = randomSeed

    def createFolds(self, data: InputOutputData, numSplits: int) -> List[Tuple[Sequence[int], Sequence[int]]]:
        numDataPoints = len(data)
        numTestPoints = numDataPoints // numSplits
        if self.shuffle:
            indices = np.random.RandomState(self.randomSeed).permutation(numDataPoints)
        else:
            indices = list(range(numDataPoints))
        result = []
        for i in range(numSplits):
            testStartIdx = i * numTestPoints
            testEndIdx = testStartIdx + numTestPoints
            testIndices = indices[testStartIdx:testEndIdx]
            trainIndices = np.concatenate((indices[:testStartIdx], indices[testEndIdx:]))
            result.append((trainIndices, testIndices))
        return result


class VectorModelCrossValidatorParams:
    def __init__(self, folds: int = 5, splitter: CrossValidationSplitter = None, returnTrainedModels=False,
            evaluatorParams: Union[VectorRegressionModelEvaluatorParams, VectorClassificationModelEvaluatorParams, dict] = None,
            defaultSplitterRandomSeed=42, defaultSplitterShuffle=True):
        """
        :param folds: the number of folds
        :param splitter: the splitter to use in order to generate the folds; if None, use default split (using parameters randomSeed
            and shuffle above)
        :param returnTrainedModels: whether to create a copy of the model for each fold and return each of the models
            (requires that models can be deep-copied); if False, the model that is passed to evalModel is fitted several times
        :param evaluatorParams: keyword parameters with which to instantiate model evaluators
        :param defaultSplitterRandomSeed: [if splitter is None] the random seed to use for splits
        :param defaultSplitterShuffle: [if splitter is None] whether to shuffle the data (using randomSeed) before creating the folds
        """
        self.folds = folds
        self.evaluatorParams = evaluatorParams
        self.returnTrainedModels = returnTrainedModels
        if splitter is None:
            splitter = CrossValidationSplitterDefault(shuffle=defaultSplitterShuffle, randomSeed=defaultSplitterRandomSeed)
        self.splitter = splitter

    @classmethod
    def fromKwArgsOldParams(cls, folds: int = 5, randomSeed=42, returnTrainedModels=False,
            evaluatorParams: Union[VectorModelEvaluatorParams, Dict[str, Any]] = None,
            shuffle=True, splitter: CrossValidationSplitter = None):
        return cls(folds=folds, splitter=splitter, returnTrainedModels=returnTrainedModels, evaluatorParams=evaluatorParams,
            defaultSplitterRandomSeed=randomSeed, defaultSplitterShuffle=shuffle)

    @classmethod
    def fromDictOrInstance(cls, params: Union[dict, "VectorModelCrossValidatorParams"]) -> "VectorModelCrossValidatorParams":
        if type(params) == dict:
            return cls.fromKwArgsOldParams(**params)
        elif isinstance(params, VectorModelCrossValidatorParams):
            return params
        else:
            raise ValueError(params)

    @classmethod
    def fromEitherDictOrInstance(cls, dictParams: dict, params: "VectorModelCrossValidatorParams"):
        if params is not None and len(dictParams) > 0:
            raise ValueError("Cannot provide both params instance and dictionary of keyword arguments")
        if params is None:
            params = cls.fromKwArgsOldParams(**dictParams)
        return params


class VectorModelCrossValidator(MetricsDictProvider, Generic[TCrossValData], ABC):
    def __init__(self, data: InputOutputData, params: Union[VectorModelCrossValidatorParams, dict] = None, **kwArgsOldParams):
        """
        :param data: the data set
        :param params: parameters
        :param kwArgsOldParams: keyword arguments for old-style specification of parameters (for backward compatibility)
        """
        self.params = VectorModelCrossValidatorParams.fromEitherDictOrInstance(kwArgsOldParams, params)
        self.modelEvaluators = []
        for trainIndices, testIndices in self.params.splitter.createFolds(data, self.params.folds):
            self.modelEvaluators.append(self._createModelEvaluator(data.filterIndices(trainIndices), data.filterIndices(testIndices)))

    @staticmethod
    def forModel(model: VectorModel, data: InputOutputData, folds=5, **kwargs) \
            -> Union["VectorClassificationModelCrossValidator", "VectorRegressionModelCrossValidator"]:
        cons = VectorRegressionModelCrossValidator if model.isRegressionModel() else VectorClassificationModelCrossValidator
        return cons(data, folds=folds, **kwargs)

    @abstractmethod
    def _createModelEvaluator(self, trainingData: InputOutputData, testData: InputOutputData) -> VectorModelEvaluator:
        pass

    @abstractmethod
    def _createResultData(self, trainedModels, evalDataList, testIndicesList, predictedVarNames) -> TCrossValData:
        pass

    def evalModel(self, model: VectorModel):
        trainedModels = [] if self.params.returnTrainedModels else None
        evalDataList = []
        testIndicesList = []
        predictedVarNames = None
        for i, evaluator in enumerate(self.modelEvaluators, start=1):
            log.info(f"Training and evaluating model with fold {i}/{len(self.modelEvaluators)} ...")
            modelToFit: VectorModel = copy.deepcopy(model) if self.params.returnTrainedModels else model
            evaluator.fitModel(modelToFit)
            evalData = evaluator.evalModel(modelToFit)
            if predictedVarNames is None:
                predictedVarNames = evalData.predictedVarNames
            if self.params.returnTrainedModels:
                trainedModels.append(modelToFit)
            for predictedVarName in predictedVarNames:
                log.info(f"Evaluation result for {predictedVarName}, fold {i}/{len(self.modelEvaluators)}: {evalData.getEvalStats(predictedVarName=predictedVarName)}")
            evalDataList.append(evalData)
            testIndicesList.append(evaluator.testData.outputs.index)
        return self._createResultData(trainedModels, evalDataList, testIndicesList, predictedVarNames)

    def _computeMetrics(self, model: VectorModel, **kwargs):
        return self._computeMetricsForVarName(model, None)

    def _computeMetricsForVarName(self, model, predictedVarName: Optional[str]):
        data = self.evalModel(model)
        return data.getEvalStatsCollection(predictedVarName=predictedVarName).aggStats()

    def createMetricsDictProvider(self, predictedVarName: Optional[str]) -> MetricsDictProvider:
        """
        Creates a metrics dictionary provider, e.g. for use in hyperparameter optimisation

        :param predictedVarName: the name of the predicted variable for which to obtain evaluation metrics; may be None only
            if the model outputs but a single predicted variable
        :return: a metrics dictionary provider instance for the given variable
        """
        return MetricsDictProviderFromFunction(functools.partial(self._computeMetricsForVarName, predictedVarName=predictedVarName))


class VectorRegressionModelCrossValidationData(VectorModelCrossValidationData[VectorRegressionModel, VectorRegressionModelEvaluationData, RegressionEvalStats, RegressionEvalStatsCollection]):
    def _createEvalStatsCollection(self, l: List[RegressionEvalStats]) -> RegressionEvalStatsCollection:
        return RegressionEvalStatsCollection(l)


class VectorRegressionModelCrossValidator(VectorModelCrossValidator[VectorRegressionModelCrossValidationData]):
    def _createModelEvaluator(self, trainingData: InputOutputData, testData: InputOutputData) -> VectorRegressionModelEvaluator:
        evaluatorParams = VectorRegressionModelEvaluatorParams.fromDictOrInstance(self.params.evaluatorParams)
        return VectorRegressionModelEvaluator(trainingData, testData=testData, params=evaluatorParams)

    def _createResultData(self, trainedModels, evalDataList, testIndicesList, predictedVarNames) -> VectorRegressionModelCrossValidationData:
        return VectorRegressionModelCrossValidationData(trainedModels, evalDataList, predictedVarNames, testIndicesList)


class VectorClassificationModelCrossValidationData(VectorModelCrossValidationData[VectorClassificationModel, VectorClassificationModelEvaluationData, ClassificationEvalStats, ClassificationEvalStatsCollection]):
    def _createEvalStatsCollection(self, l: List[ClassificationEvalStats]) -> ClassificationEvalStatsCollection:
        return ClassificationEvalStatsCollection(l)


class VectorClassificationModelCrossValidator(VectorModelCrossValidator[VectorClassificationModelCrossValidationData]):
    def _createModelEvaluator(self, trainingData: InputOutputData, testData: InputOutputData):
        evaluatorParams = VectorClassificationModelEvaluatorParams.fromDictOrInstance(self.params.evaluatorParams)
        return VectorClassificationModelEvaluator(trainingData, testData=testData, params=evaluatorParams)

    def _createResultData(self, trainedModels, evalDataList, testIndicesList, predictedVarNames) -> VectorClassificationModelCrossValidationData:
        return VectorClassificationModelCrossValidationData(trainedModels, evalDataList, predictedVarNames, testIndicesList)



import collections
from typing import Hashable, Dict

from .string import ToStringMixin


class RelativeFrequencyCounter(ToStringMixin):
    """
    Counts the absolute and relative frequency of an event
    """
    def __init__(self):
        self.numTotal = 0
        self.numRelevant = 0

    def count(self, isRelevantEvent) -> None:
        """
        Adds to the the count.
        The nominator is incremented only if we are counting a relevant event.
        The denominator is always incremented.

        :param isRelevantEvent: whether we are counting a relevant event
        """
        self.numTotal += 1
        if isRelevantEvent:
            self.numRelevant += 1

    def _toStringObjectInfo(self):
        info = f"{self.numRelevant}/{self.numTotal}"
        if self.numTotal > 0:
            info += f", {100 * self.numRelevant / self.numTotal:.2f}%"
        return info

    def add(self, relativeFrequencyCounter: __qualname__) -> None:
        """
        Adds the counts of the given counter to this object

        :param relativeFrequencyCounter: the counter whose data to add
        """
        self.numTotal += relativeFrequencyCounter.numTotal
        self.numRelevant += relativeFrequencyCounter.numRelevant

    def getRelativeFrequency(self) -> float:
        return self.numRelevant / self.numTotal


class DistributionCounter(ToStringMixin):
    """
    Supports the counting of the frequencies with which (mutually exclusive) events occur
    """
    def __init__(self):
        self.counts = collections.defaultdict(lambda: 0)
        self.totalCount = 0

    def count(self, event: Hashable) -> None:
        """
        Increments the count of the given event

        :param event: the event/key whose count to increment, which must be hashable
        """
        self.totalCount += 1
        self.counts[event] += 1

    def getDistribution(self) -> Dict[Hashable, float]:
        """
        :return: a dictionary mapping events (as previously passed to count) to their relative frequencies
        """
        return {k: v/self.totalCount for k, v in self.counts.items()}

    def _toStringObjectInfo(self):
        return ", ".join([f"{str(k)}: {v} ({v/self.totalCount:.3f})" for k, v in self.counts.items()])


class WeightedMean:
    """
    Computes a weighted mean of values
    """
    def __init__(self):
        self.weightedValueSum = 0
        self.weightSum = 0

    def add(self, value, weight) -> None:
        """
        Adds the given value with the the given weight to the calculation

        :param value: the value
        :param weight: the weight with which to consider tha value
        """
        self.weightedValueSum += value * weight
        self.weightSum += weight

    def getWeightedMean(self):
        """
        :return: the weighted mean of all values that have been added
        """
        return self.weightedValueSum / self.weightSum


from . import columngen
from . import data_transformation
from . import featuregen
from . import hyperopt
from . import local_search
from . import naive_bayes
from . import nearest_neighbors
from . import sklearn
from . import util
from .data import InputOutputData
from .data_transformation import DataFrameTransformer, RuleBasedDataFrameTransformer
from .ensemble import AveragingVectorRegressionModel
from .evaluation.eval_stats import eval_stats_classification, eval_stats_regression
from .normalisation import NormalisationMode
from .tensor_model import TensorToTensorRegressionModel, TensorToScalarRegressionModel, \
    TensorToTensorClassificationModel, TensorToScalarClassificationModel
from .vector_model import VectorModelBase, VectorModel, VectorRegressionModel, VectorClassificationModel

__version__ = "0.1.7"

# The following submodules are not imported by default to avoid necessarily requiring their dependencies:
# tensorflow
# torch
# lightgbm
# catboost
# geoanalytics
# xgboost



"""
This module contains methods and classes that facilitate evaluation of different types of models. The suggested
workflow for evaluation is to use these higher-level functionalities instead of instantiating
the evaluation classes directly.
"""
# TODO: provide a notebook (and possibly an rst file) that illustrates standard evaluation scenarios and at the same
#  time serves as an integration test

import logging
from abc import ABC, abstractmethod
from typing import Tuple, Dict, Any, Union, Generic, TypeVar, Optional, Sequence, Callable

import matplotlib.figure
import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns

from .crossval import VectorModelCrossValidationData, VectorRegressionModelCrossValidationData, \
    VectorClassificationModelCrossValidationData, \
    VectorClassificationModelCrossValidator, VectorRegressionModelCrossValidator, VectorModelCrossValidator, VectorModelCrossValidatorParams
from .eval_stats.eval_stats_base import EvalStats, EvalStatsCollection
from .eval_stats.eval_stats_classification import ClassificationEvalStats
from .eval_stats.eval_stats_regression import RegressionEvalStats
from .evaluator import VectorModelEvaluator, VectorModelEvaluationData, VectorRegressionModelEvaluator, \
    VectorRegressionModelEvaluationData, VectorClassificationModelEvaluator, VectorClassificationModelEvaluationData, \
    VectorRegressionModelEvaluatorParams, VectorClassificationModelEvaluatorParams, VectorModelEvaluatorParams
from ..data import InputOutputData
from ..util.io import ResultWriter
from ..util.string import prettyStringRepr
from ..vector_model import VectorClassificationModel, VectorRegressionModel, VectorModel

log = logging.getLogger(__name__)

TModel = TypeVar("TModel", bound=VectorModel)
TEvalStats = TypeVar("TEvalStats", bound=EvalStats)
TEvalStatsCollection = TypeVar("TEvalStatsCollection", bound=EvalStatsCollection)
TEvaluator = TypeVar("TEvaluator", bound=VectorModelEvaluator)
TCrossValidator = TypeVar("TCrossValidator", bound=VectorModelCrossValidator)
TEvalData = TypeVar("TEvalData", bound=VectorModelEvaluationData)
TCrossValData = TypeVar("TCrossValData", bound=VectorModelCrossValidationData)


def _isRegression(model: Optional[VectorModel], isRegression: Optional[bool]) -> bool:
    if model is None and isRegression is None or (model is not None and isRegression is not None):
        raise ValueError("One of the two parameters have to be passed: model or isRegression")

    if isRegression is None:
        model: VectorModel
        return model.isRegressionModel()
    return isRegression


def createVectorModelEvaluator(data: InputOutputData, model: VectorModel = None,
        isRegression: bool = None, params: Union[VectorModelEvaluatorParams, Dict[str, Any]] = None, **kwargs) \
            -> Union[VectorRegressionModelEvaluator, VectorClassificationModelEvaluator]:
    if params is not None and len(kwargs) > 0:
        raise ValueError("Provide either params or keyword arguments")
    if params is None:
        params = kwargs
    regression = _isRegression(model, isRegression)
    if regression:
        params = VectorRegressionModelEvaluatorParams.fromDictOrInstance(params)
    else:
        params = VectorClassificationModelEvaluatorParams.fromDictOrInstance(params)
    cons = VectorRegressionModelEvaluator if regression else VectorClassificationModelEvaluator
    return cons(data, params=params)


def createVectorModelCrossValidator(data: InputOutputData, model: VectorModel = None,
        isRegression: bool = None,
        params: Union[VectorModelCrossValidatorParams, Dict[str, Any]] = None,
        **kwArgsOldParams) -> Union[VectorClassificationModelCrossValidator, VectorRegressionModelCrossValidator]:
    if params is not None:
        params = VectorModelCrossValidatorParams.fromDictOrInstance(params)
    params = VectorModelCrossValidatorParams.fromEitherDictOrInstance(kwArgsOldParams, params)
    cons = VectorRegressionModelCrossValidator if _isRegression(model, isRegression) else VectorClassificationModelCrossValidator
    return cons(data, params=params)


def createEvaluationUtil(data: InputOutputData, model: VectorModel = None, isRegression: bool = None,
        evaluatorParams: Optional[Dict[str, Any]] = None,
        crossValidatorParams: Optional[Dict[str, Any]] = None) \
            -> Union["ClassificationEvaluationUtil", "RegressionEvaluationUtil"]:
    cons = RegressionEvaluationUtil if _isRegression(model, isRegression) else ClassificationEvaluationUtil
    return cons(data, evaluatorParams=evaluatorParams, crossValidatorParams=crossValidatorParams)


def evalModelViaEvaluator(model: TModel, inputOutputData: InputOutputData, testFraction=0.2,
        plotTargetDistribution=False, computeProbabilities=True, normalizePlots=True, randomSeed=60) -> TEvalData:
    """
    Evaluates the given model via a simple evaluation mechanism that uses a single split

    :param model: the model to evaluate
    :param inputOutputData: data on which to evaluate
    :param testFraction: the fraction of the data to test on
    :param plotTargetDistribution: whether to plot the target values distribution in the entire dataset
    :param computeProbabilities: only relevant if the model is a classifier
    :param normalizePlots: whether to normalize plotted distributions such that the sum/integrate to 1
    :param randomSeed:

    :return: the evaluation data
    """
    if plotTargetDistribution:
        title = "Distribution of target values in entire dataset"
        fig = plt.figure(title)

        outputDistributionSeries = inputOutputData.outputs.iloc[:, 0]
        log.info(f"Description of target column in training set: \n{outputDistributionSeries.describe()}")
        if not model.isRegressionModel():
            outputDistributionSeries = outputDistributionSeries.value_counts(normalize=normalizePlots)
            ax = sns.barplot(outputDistributionSeries.index, outputDistributionSeries.values)
            ax.set_ylabel("%")
        else:
            ax = sns.distplot(outputDistributionSeries)
            ax.set_ylabel("Probability density")
        ax.set_title(title)
        ax.set_xlabel("target value")
        fig.show()

    if model.isRegressionModel():
        evaluatorParams = dict(testFraction=testFraction, randomSeed=randomSeed)
    else:
        evaluatorParams = dict(testFraction=testFraction, computeProbabilities=computeProbabilities, randomSeed=randomSeed)
    ev = createEvaluationUtil(inputOutputData, model=model, evaluatorParams=evaluatorParams)
    return ev.performSimpleEvaluation(model, showPlots=True, logResults=True)


class EvaluationUtil(ABC, Generic[TModel, TEvaluator, TEvalData, TCrossValidator, TCrossValData, TEvalStats]):
    """
    Utility class for the evaluation of models based on a dataset
    """
    def __init__(self, inputOutputData: InputOutputData,
            evaluatorParams: Optional[Union[VectorRegressionModelEvaluatorParams, VectorClassificationModelEvaluatorParams, Dict[str, Any]]] = None,
            crossValidatorParams: Optional[Union[VectorModelCrossValidatorParams, Dict[str, Any]]] = None):
        """
        :param inputOutputData: the data set to use for evaluation
        :param evaluatorParams: parameters with which to instantiate evaluators
        :param crossValidatorParams: parameters with which to instantiate cross-validators
        """
        if evaluatorParams is None:
            evaluatorParams = dict(testFraction=0.2)
        if crossValidatorParams is None:
            crossValidatorParams = VectorModelCrossValidatorParams(folds=5)
        self.evaluatorParams = evaluatorParams
        self.crossValidatorParams = crossValidatorParams
        self.inputOutputData = inputOutputData

    class ResultCollector:
        def __init__(self, showPlots: bool = True, resultWriter: Optional[ResultWriter] = None):
            self.showPlots = showPlots
            self.resultWriter = resultWriter

        def addFigure(self, name, fig: matplotlib.figure.Figure):
            if self.resultWriter is not None:
                self.resultWriter.writeFigure(name, fig, closeFigure=not self.showPlots)

        def child(self, addedFilenamePrefix):
            resultWriter = self.resultWriter
            if resultWriter:
                resultWriter = resultWriter.childWithAddedPrefix(addedFilenamePrefix)
            return self.__class__(showPlots=self.showPlots, resultWriter=resultWriter)

    def createEvaluator(self, model: TModel = None, isRegression: bool = None) -> TEvaluator:
        """
        Creates an evaluator holding the current input-output data

        :param model: the model for which to create an evaluator (just for reading off regression or classification,
            the resulting evaluator will work on other models as well)
        :param isRegression: whether to create a regression model evaluator. Either this or model have to be specified
        :return: an evaluator
        """
        return createVectorModelEvaluator(self.inputOutputData, model=model, isRegression=isRegression, params=self.evaluatorParams)

    def createCrossValidator(self, model: TModel = None, isRegression: bool = None) -> TCrossValidator:
        """
        Creates a cross-validator holding the current input-output data

        :param model: the model for which to create a cross-validator (just for reading off regression or classification,
            the resulting evaluator will work on other models as well)
        :param isRegression: whether to create a regression model cross-validator. Either this or model have to be specified
        :return: an evaluator
        """
        return createVectorModelCrossValidator(self.inputOutputData, model=model, isRegression=isRegression, params=self.crossValidatorParams)

    def performSimpleEvaluation(self, model: TModel, createPlots=True, showPlots=False, logResults=True, resultWriter: ResultWriter = None,
            additionalEvaluationOnTrainingData=False, fitModel=True, writeEvalStats=False) -> TEvalData:
        if showPlots and not createPlots:
            raise ValueError("showPlots=True requires createPlots=True")
        resultWriter = self._resultWriterForModel(resultWriter, model)
        evaluator = self.createEvaluator(model)
        log.info(f"Evaluating {model} via {evaluator}")
        if fitModel:
            evaluator.fitModel(model)

        def gatherResults(evalResultData: VectorModelEvaluationData, resultWriter, subtitlePrefix=""):
            strEvalResults = ""
            for predictedVarName in evalResultData.predictedVarNames:
                evalStats = evalResultData.getEvalStats(predictedVarName)
                strEvalResult = str(evalStats)
                if logResults:
                    log.info(f"{subtitlePrefix}Evaluation results for {predictedVarName}: {strEvalResult}")
                strEvalResults += predictedVarName + ": " + strEvalResult + "\n"
                if writeEvalStats and resultWriter is not None:
                    resultWriter.writePickle(f"eval-stats-{predictedVarName}", evalStats)
            strEvalResults += f"\n\n{prettyStringRepr(model)}"
            if resultWriter is not None:
                resultWriter.writeTextFile("evaluator-results", strEvalResults)
            if createPlots:
                self.createPlots(evalResultData, showPlots=showPlots, resultWriter=resultWriter, subtitlePrefix=subtitlePrefix)

        evalResultData = evaluator.evalModel(model)
        gatherResults(evalResultData, resultWriter)
        if additionalEvaluationOnTrainingData:
            evalResultDataTrain = evaluator.evalModel(model, onTrainingData=True)
            additionalResultWriter = resultWriter.childWithAddedPrefix("onTrain-") if resultWriter is not None else None
            gatherResults(evalResultDataTrain, additionalResultWriter, subtitlePrefix="[onTrain] ")

        return evalResultData

    @staticmethod
    def _resultWriterForModel(resultWriter: Optional[ResultWriter], model: TModel) -> Optional[ResultWriter]:
        if resultWriter is None:
            return None
        return resultWriter.childWithAddedPrefix(model.getName() + "-")

    def performCrossValidation(self, model: TModel, showPlots=False, logResults=True, resultWriter: Optional[ResultWriter] = None) -> TCrossValData:
        """
        Evaluates the given model via cross-validation

        :param model: the model to evaluate
        :param showPlots: whether to show plots that visualise evaluation results (combining all folds)
        :param logResults: whether to log evaluation results
        :param resultWriter: a writer with which to store text files and plots. The evaluated model's name is added to each filename
            automatically
        :return: cross-validation result data
        """
        resultWriter = self._resultWriterForModel(resultWriter, model)
        crossValidator = self.createCrossValidator(model)
        crossValidationData = crossValidator.evalModel(model)
        aggStatsByVar = {varName: crossValidationData.getEvalStatsCollection(predictedVarName=varName).aggStats()
                for varName in crossValidationData.predictedVarNames}
        strEvalResults = str(pd.DataFrame.from_dict(aggStatsByVar, orient="index"))
        if logResults:
            log.info(f"Cross-validation results:\n{strEvalResults}")
        if resultWriter is not None:
            resultWriter.writeTextFile("crossval-results", strEvalResults)
        self.createPlots(crossValidationData, showPlots=showPlots, resultWriter=resultWriter)
        return crossValidationData

    def compareModels(self, models: Sequence[TModel], resultWriter: Optional[ResultWriter] = None, useCrossValidation=False,
            fitModels=True, writeIndividualResults=True, sortColumn: Optional[str] = None, sortAscending: bool = True) -> pd.DataFrame:
        """
        Compares several models via simple evaluation or cross-validation

        :param models: the models to compare
        :param resultWriter: a writer with which to store results of the comparison
        :param useCrossValidation: whether to use cross-validation in order to evaluate models; if False, use a simple evaluation
            on test data (single split)
        :param fitModels: whether to fit models before evaluating them; this can only be False if useCrossValidation=False
        :param writeIndividualResults: whether to write results files on each individual model (in addition to the comparison
            summary)
        :param sortColumn: column/metric name by which to sort
        :param sortAscending: whether to sort in ascending order
        :return: a data frame containing evaluation metrics on all models
        """
        statsList = []
        for model in models:
            if useCrossValidation:
                if not fitModels:
                    raise ValueError("Cross-validation necessitates that models be retrained; got fitModels=False")
                crossValidationResult = self.performCrossValidation(model, resultWriter=resultWriter if writeIndividualResults else None)
                statsDict = crossValidationResult.getEvalStatsCollection().aggStats()
            else:
                evalStats: EvalStats = self.performSimpleEvaluation(model, resultWriter=resultWriter if writeIndividualResults else None,
                    fitModel=fitModels).getEvalStats()
                statsDict = evalStats.getAll()
            statsDict["modelName"] = model.getName()
            statsList.append(statsDict)
        resultsDF = pd.DataFrame(statsList).set_index("modelName")
        if sortColumn is not None:
            if sortColumn not in resultsDF.columns:
                log.warning(f"Requested sort column '{sortColumn}' not in list of columns {list(resultsDF.columns)}")
            else:
                resultsDF.sort_values(sortColumn, ascending=sortAscending, inplace=True)
        strResults = f"Model comparison results:\n{resultsDF.to_string()}"
        log.info(strResults)
        if resultWriter is not None:
            suffix = "crossval" if useCrossValidation else "simple-eval"
            strResults += "\n\n" + "\n\n".join([f"{model.getName()} = {str(model)}" for model in models])
            resultWriter.writeTextFile(f"model-comparison-results-{suffix}", strResults)
        return resultsDF

    def compareModelsCrossValidation(self, models: Sequence[TModel], resultWriter: Optional[ResultWriter] = None) -> pd.DataFrame:
        """
        Compares several models via cross-validation

        :param models: the models to compare
        :param resultWriter: a writer with which to store results of the comparison
        :return: a data frame containing evaluation metrics on all models
        """
        return self.compareModels(models, resultWriter=resultWriter, useCrossValidation=True)

    def createPlots(self, data: Union[TEvalData, TCrossValData], showPlots=True, resultWriter: Optional[ResultWriter] = None, subtitlePrefix: str = ""):
        """
        Creates default plots that visualise the results in the given evaluation data

        :param data: the evaluation data for which to create the default plots
        :param showPlots: whether to show plots
        :param resultWriter: if not None, plots will be written using this writer
        :param subtitlePrefix: a prefix to add to the subtitle (which itself is the model name)
        """
        if not showPlots and resultWriter is None:
            return
        resultCollector = self.ResultCollector(showPlots=showPlots, resultWriter=resultWriter)
        self._createPlots(data, resultCollector, subtitle=subtitlePrefix + data.modelName)

    def _createPlots(self, data: Union[TEvalData, TCrossValData], resultCollector: ResultCollector, subtitle=None):

        def createPlots(predVarName, rc, subt):
            if isinstance(data, VectorModelCrossValidationData):
                evalStats = data.getEvalStatsCollection(predictedVarName=predVarName).getGlobalStats()
            elif isinstance(data, VectorModelEvaluationData):
                evalStats = data.getEvalStats(predictedVarName=predVarName)
            else:
                raise ValueError(f"Unexpected argument: data={data}")
            return self._createEvalStatsPlots(evalStats, rc, subtitle=subt)

        predictedVarNames = data.predictedVarNames
        if len(predictedVarNames) == 1:
            createPlots(predictedVarNames[0], resultCollector, subtitle)
        else:
            for predictedVarName in predictedVarNames:
                createPlots(predictedVarName, resultCollector.child(predictedVarName+"-"), f"{predictedVarName}, {subtitle}")

    @abstractmethod
    def _createEvalStatsPlots(self, evalStats: TEvalStats, resultCollector: ResultCollector, subtitle=None):
        """
        :param evalStats: the evaluation results for which to create plots
        :param resultCollector: the collector to which all plots are to be passed
        :param subtitle: the subtitle to use for generated plots (if any)
        """
        pass


class RegressionEvaluationUtil(EvaluationUtil[VectorRegressionModel, VectorRegressionModelEvaluator, VectorRegressionModelEvaluationData, VectorRegressionModelCrossValidator, VectorRegressionModelCrossValidationData, RegressionEvalStats]):
    def _createEvalStatsPlots(self, evalStats: RegressionEvalStats, resultCollector: EvaluationUtil.ResultCollector, subtitle=None):
        resultCollector.addFigure("error-dist", evalStats.plotErrorDistribution(titleAdd=subtitle))
        resultCollector.addFigure("heatmap-gt-pred", evalStats.plotHeatmapGroundTruthPredictions(titleAdd=subtitle))
        resultCollector.addFigure("scatter-gt-pred", evalStats.plotScatterGroundTruthPredictions(titleAdd=subtitle))


class ClassificationEvaluationUtil(EvaluationUtil[VectorClassificationModel, VectorClassificationModelEvaluator, VectorClassificationModelEvaluationData, VectorClassificationModelCrossValidator, VectorClassificationModelCrossValidationData, ClassificationEvalStats]):
    def _createEvalStatsPlots(self, evalStats: ClassificationEvalStats, resultCollector: EvaluationUtil.ResultCollector, subtitle=None):
        resultCollector.addFigure("confusion-matrix", evalStats.plotConfusionMatrix(titleAdd=subtitle))


class MultiDataEvaluationUtil:
    def __init__(self, inputOutputDataDict: Dict[str, InputOutputData], keyName: str = "dataset"):
        """
        :param inputOutputDataDict: a dictionary mapping from names to the data sets with which to evaluate models
        :param keyName: a name for the key value used in inputOutputDataDict
        """
        self.inputOutputDataDict = inputOutputDataDict
        self.keyName = keyName

    def compareModelsCrossValidation(self, modelFactories: Sequence[Callable[[], VectorModel]],
            resultWriter: Optional[ResultWriter] = None, writePerDatasetResults=True,
            crossValidatorParams: Optional[Dict[str, Any]] = None, columnNameForModelRanking: str = None, rankMax=True) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """
        :param modelFactories: a sequence of factory functions for the creation of models to evaluate
        :param resultWriter: a writer with which to store results
        :param writePerDatasetResults: whether to use resultWriter (if not None) in order to generate detailed results for each
            dataset in a subdirectory named according to the name of the dataset
        :param crossValidatorParams: parameters to use for the instantiation of cross-validators
        :param columnNameForModelRanking: column name to use for ranking models
        :param rankMax: if true, use max for ranking, else min
        :return: a pair of data frames (allDF, meanDF) where allDF contains all the individual cross-validation results
            for every dataset and meanDF contains one row for each model with results averaged across datasets
        """
        allResults = pd.DataFrame()
        for key, inputOutputData in self.inputOutputDataDict.items():
            log.info(f"Evaluating models for {key}")
            models = [f() for f in modelFactories]
            modelsAreRegression = [model.isRegressionModel() for model in models]
            if all(modelsAreRegression):
                isRegression = True
            elif not any(modelsAreRegression):
                isRegression = False
            else:
                raise ValueError("The models have to be either all regression models or all classification, not a mixture")
            ev = createEvaluationUtil(inputOutputData, isRegression=isRegression, crossValidatorParams=crossValidatorParams)
            childResultWriter = resultWriter.childForSubdirectory(key) if writePerDatasetResults else None
            df = ev.compareModelsCrossValidation(models, resultWriter=childResultWriter)
            df[self.keyName] = key
            df["modelName"] = df.index
            if columnNameForModelRanking is not None:
                if columnNameForModelRanking not in df.columns:
                    raise ValueError(f"Rank metric {columnNameForModelRanking} not contained in columns {df.columns}")
                df["bestModel"] = 0
                if rankMax:
                    df["bestModel"].loc[df[columnNameForModelRanking].idxmax()] = 1
                else:
                    df["bestModel"].loc[df[columnNameForModelRanking].idxmin()] = 1
            df = df.reset_index(drop=True)
            allResults = pd.concat((allResults, df))
        strAllResults = f"All results:\n{allResults.to_string()}"
        log.info(strAllResults)
        meanResults = allResults.groupby("modelName").mean()
        strMeanResults = f"Mean results:\n{meanResults.to_string()}"
        log.info(strMeanResults)
        if resultWriter is not None:
            resultWriter.writeTextFile("model-comparison-results", strMeanResults + "\n\n" + strAllResults)
        return allResults, meanResults



from bisect import bisect_right, bisect_left
from typing import Optional, TypeVar, Sequence, Any, List

T = TypeVar("T")


def concatSequences(seqs: Sequence[Sequence[T]]) -> List[T]:
    result = []
    for s in seqs:
        result.extend(s)
    return result


def getFirstDuplicate(seq: Sequence[T]) -> Optional[T]:
    """
    Returns the first duplicate in a sequence or None

    :param seq: a sequence of hashable elements
    :return:
    """
    setOfElems = set()
    for elem in seq:
        if elem in setOfElems:
            return elem
        setOfElems.add(elem)


def floorIndex(arr, value) -> Optional[int]:
    """
    Finds the rightmost/largest index in a sorted array where the value is less than or equal to the given value

    :param arr: the sorted array of values
    :param value: the value to search for
    :return: the index or None if there is no such index
    """
    idx = bisect_right(arr, value)
    if idx:
        return idx - 1
    return None


def ceilIndex(arr, value) -> Optional[int]:
    """
    finds the leftmost/lowest index in a sorted array where the value is greater than or equal to the given value

    :param arr: the sorted array of values
    :param value: the value to search for
    :return: the index or None if there is no such index
    """
    idx = bisect_left(arr, value)
    if idx != len(arr):
        return idx
    return None


def closestIndex(arr, value) -> Optional[int]:
    """
    Finds the index in the given array where the value is closest to the given value.
    If two subsequent values have the same distance, the smaller index is returned.

    :param arr: the array to search in
    :param value: the value to search for
    :return: the index or None if the array is empty
    """
    length = len(arr)
    if length == 0:
        return None
    floorIdx = floorIndex(arr, value)
    if floorIdx is None:
        return 0
    ceilIdx = floorIdx + 1
    if ceilIdx >= length:
        return floorIdx
    floorValue = arr[floorIdx]
    ceilValue = arr[floorIdx]
    floorDist = abs(floorValue - value)
    ceilDist = abs(ceilValue - value)
    if floorDist <= ceilDist:
        return floorIdx
    else:
        return ceilIdx


def ceilValue(keys, keyValue, values=None) -> Optional[Any]:
    """
    For a sorted array of keys (and a an array of corresponding values),
    returns the corresponding value (values[i]) for the lowest index i where keys[i] >= keyValue

    :param keys: the sorted array of keys for in which to search for the value closest to `keyValue`
    :param keyValue: the key to search for
    :param values: the array from which to retrieve the value; if None, use `keys`
    :return: the value or None if no such value exists
    """
    idx = ceilIndex(keys, keyValue)
    if idx is None:
        return None
    values = values if values is not None else keys
    return values[idx]


def floorValue(keys, keyValue, values=None, fallbackFirst=False) -> Optional[Any]:
    """
    For a sorted array of keys (and an array of corresponding values),
    returns the corresponding value (values[i]) for the largest index i in keys where keys[i] <= keyValue.

    :param keys: the sorted array of keys in which to perform the lookup
    :param keyValue: the key for which to perform the lookup
    :param values: the sorted array of values; if None, use keys as values
    :param fallbackFirst: if True, then return the first value instead of None for the case where no floor value exists
    :return: the floor value
    """
    idx = floorIndex(keys, keyValue)
    values = values if values is not None else keys
    if idx is None:
        if fallbackFirst:
            return values[0]
        return None
    return values[idx]


def closestValue(keys, keyValue, values=None) -> Optional[Any]:
    """
    For a sorted array of keys (and an array of corresponding values),
    finds the value at the index where the key is closest to the given key value.
    If two subsequent values are equally close, the value at the smaller index is returned.

    :param keys:
    :param value:
    :return:
    """
    idx = closestIndex(keys, keyValue)
    if idx is None:
        return None
    if values is None:
        values = keys
    return values[idx]


def valueSliceInner(keys, lowerBoundKey, upperBoundKey, values=None):
    """
    For a sorted array of keys (and a an array of corresponding values),
    finds indices i, j such that i is the lowest key where keys[i] >= lowerBoundKey and
    j is the largest key where keys[j] <= upperBoundKey,
    and returns the corresponding slice of values values[i:j+1],
    i.e. the slice will not include the bounds keys if they are not present in the keys array.

    :param keys: the sorted array of key values
    :param lowerBoundKey: the key value defining the lower bound
    :param upperBoundKey: the key value defining the upper bound
    :param values: the sorted array of values; if None, use keys
    :return: the corresponding slice of `values`
    """
    if values is None:
        values = keys
    elif len(values) != len(keys):
        raise ValueError("Length of values must match length of keys")
    if lowerBoundKey > upperBoundKey:
        raise ValueError(f"Invalid bounds: {lowerBoundKey} to {upperBoundKey}")
    firstIdx = ceilIndex(keys, lowerBoundKey)
    lastIdx = floorIndex(keys, upperBoundKey)
    if firstIdx is None or lastIdx is None:
        return values[0:0]  # empty slice
    return values[firstIdx:lastIdx+1]


def valueSliceOuter(keys, lowerBoundKey, upperBoundKey, values=None, fallbackBounds=True):
    """
    For a sorted array of keys and an array of corresponding values,
    finds indices i, j such that i is the largest key where keys[i] <= lowerBoundKey and
    j is the lowest key where keys[j] <= upperBoundKey,
    and returns the corresponding slice of values values[i:j+1].
    If such indices do not exists and fallbackBounds==True, the array bounds are used (i.e. 0 or len-1).
    Given the indices i and j, the function returns values[i:j+1].
    This is an outer slice which is the smallest slice that definitely contains two given bounds
    (for fallbackBounds==False).

    :param keys: the sorted array of key values
    :param lowerBoundKey: the key value defining the lower bound
    :param upperBoundKey: the key value defining the upper bound
    :param values: the sorted array of values; if None, use keys
    :return: the corresponding slice of `values`
    """
    if values is None:
        values = keys
    elif len(values) != len(keys):
        raise ValueError("Length of values must match length of keys")
    if lowerBoundKey > upperBoundKey:
        raise ValueError(f"Invalid bounds: {lowerBoundKey} to {upperBoundKey}")
    firstIdx = floorIndex(keys, lowerBoundKey)
    lastIdx = ceilIndex(keys, upperBoundKey)
    if firstIdx is None:
        if fallbackBounds:
            firstIdx = 0
        else:
            raise Exception(f"Bounds would exceed start of array (lowerBoundKey={lowerBoundKey}, lowest key={keys[0]})")
    if lastIdx is None:
        if fallbackBounds:
            lastIdx = len(keys) - 1
        else:
            raise Exception(f"Bounds would exceed end of array (upperBoundKey={upperBoundKey}, largest key={keys[-1]})")
    return values[firstIdx:lastIdx+1]



import numpy as np
from pytorch_lightning import Trainer, LightningModule
from torch import tensor

from .. import tensor_model as tm
from ..data import InputOutputArrays, DataSplitterFractional


def _fitModelWithTrainer(model: LightningModule, trainer: Trainer, inputOutputData,
                         batchSize: int, splitter: DataSplitterFractional = None):
    if splitter is not None:
        trainIOData, validationIOData = splitter.split(inputOutputData)
        trainDataLoader = trainIOData.toTorchDataLoader(batchSize=batchSize)
        valDataLoader = validationIOData.toTorchDataLoader(batchSize=batchSize)
    else:
        trainDataLoader = inputOutputData.toTorchDataLoader(batchSize=batchSize)
        valDataLoader = None
    trainer.fit(model, trainDataLoader, val_dataloaders=valDataLoader)


class PLWrappedModel:
    def __init__(self, model: LightningModule, trainer: Trainer, validationFraction=0.1, shuffle=True, batchSize=32):
        if not 0 <= validationFraction <= 1:
            raise ValueError(f"Invalid validationFraction: {validationFraction}. Has to be in interval [0, 1]")
        self.trainer = trainer
        self.model = model
        self.validationFraction = validationFraction
        self.shuffle = shuffle
        self.batchSize = batchSize

    def fit(self, X: np.ndarray, Y: np.ndarray):
        inputOutputData = InputOutputArrays(X, Y)
        splitter = DataSplitterFractional(1 - self.validationFraction, shuffle=self.shuffle)
        _fitModelWithTrainer(self.model, self.trainer, inputOutputData, self.batchSize, splitter=splitter)

    def predict(self, X: np.ndarray) -> np.ndarray:
        X = tensor(X)
        return self.model(X).detach().cpu().numpy()


# noinspection DuplicatedCode
class PLTensorToScalarClassificationModel(tm.TensorToScalarClassificationModel):
    def __init__(self, model: LightningModule, trainer: Trainer, validationFraction=0.1, shuffle=True, batchSize=64,
                 checkInputShape=True, checkInputColumns=True):
        super().__init__(checkInputShape=checkInputShape, checkInputColumns=checkInputColumns)
        self.wrappedModel = PLWrappedModel(model, trainer, validationFraction=validationFraction, shuffle=shuffle,
                                           batchSize=batchSize)

    def _predictProbabilitiesArray(self, X: np.ndarray) -> np.ndarray:
        return self.wrappedModel.predict(X)

    def _fitToArray(self, X: np.ndarray, Y: np.ndarray):
        self.wrappedModel.fit(X, Y)


# noinspection DuplicatedCode
class PLTensorToScalarRegressionModel(tm.TensorToScalarRegressionModel):
    def __init__(self, model: LightningModule, trainer: Trainer,  validationFraction=0.1, shuffle=True, batchSize=32,
                 checkInputShape=True, checkInputColumns=True):
        super().__init__(checkInputShape=checkInputShape, checkInputColumns=checkInputColumns)
        self.wrappedModel = PLWrappedModel(model, trainer, validationFraction=validationFraction, shuffle=shuffle,
                                           batchSize=batchSize)

    def _predictArray(self, X: np.ndarray) -> np.ndarray:
        return self.wrappedModel.predict(X)

    def _fitToArray(self, X: np.ndarray, Y: np.ndarray):
        self.wrappedModel.fit(X, Y)


# noinspection DuplicatedCode
class PLTensorToTensorClassificationModel(tm.TensorToTensorClassificationModel):
    def __init__(self, model: LightningModule, trainer: Trainer,  validationFraction=0.1, shuffle=True, batchSize=32,
                 checkInputShape=True, checkInputColumns=True):
        super().__init__(checkInputShape=checkInputShape, checkInputColumns=checkInputColumns)
        self.wrappedModel = PLWrappedModel(model, trainer, validationFraction=validationFraction, shuffle=shuffle,
                                           batchSize=batchSize)

    def _predictProbabilitiesArray(self, X: np.ndarray) -> np.ndarray:
        return self.wrappedModel.predict(X)

    def _fitToArray(self, X: np.ndarray, Y: np.ndarray):
        self.wrappedModel.fit(X, Y)


# noinspection DuplicatedCode
class PLTensorToTensorRegressionModel(tm.TensorToTensorRegressionModel):
    def __init__(self, model: LightningModule, trainer: Trainer,  validationFraction=0.1, shuffle=True, batchSize=32,
                 checkInputShape=True, checkInputColumns=True):
        super().__init__(checkInputShape=checkInputShape, checkInputColumns=checkInputColumns)
        self.wrappedModel = PLWrappedModel(model, trainer, validationFraction=validationFraction, shuffle=shuffle,
                                           batchSize=batchSize)

    def _predictArray(self, X: np.ndarray) -> np.ndarray:
        return self.wrappedModel.predict(X)

    def _fitToArray(self, X: np.ndarray, Y: np.ndarray):
        self.wrappedModel.fit(X, Y)



from .crossval import VectorClassificationModelCrossValidator, VectorRegressionModelCrossValidator, \
    VectorClassificationModelCrossValidationData, VectorRegressionModelCrossValidationData
from .eval_util import RegressionEvaluationUtil, ClassificationEvaluationUtil, MultiDataEvaluationUtil, \
    evalModelViaEvaluator, createEvaluationUtil, createVectorModelEvaluator, createVectorModelCrossValidator
from .evaluator import VectorClassificationModelEvaluator, VectorRegressionModelEvaluator, \
    VectorRegressionModelEvaluationData, VectorClassificationModelEvaluationData, \
    RuleBasedVectorClassificationModelEvaluator, RuleBasedVectorRegressionModelEvaluator

# imports required for backward compatibility
from ..data import DataSplitter, DataSplitterFractional



import functools
import logging
from typing import Optional, Callable
from time import time


log = logging.getLogger(__name__)


def profiled(func: Optional[Callable] = None, *, sampling_interval_seconds: Optional[float] = None,
        print_to_console=True, log_to_file=False, file_name: Optional[str] = None):
    """
    Function decorator for profiling the annotated/given function with the pyinstrument profiler
    :param func: the function to be profiled
    :param sampling_interval_seconds: sampling every <sampling_interval_seconds> seconds, if None use default parameter of pyintrument profiler
    :param print_to_console: if, true print profiler output to console
    :param log_to_file: if true, logs profiler output to html file
    :param file_name: optional file name for log, if None defaults to file name derived from function name
    :return: the wrapped function
    """
    from pyinstrument import Profiler

    def wrapper(_func):
        @functools.wraps(_func)
        def _wrapper(*args, **kwargs):
            _profiler = Profiler(sampling_interval_seconds) if sampling_interval_seconds is not None else Profiler()
            _profiler.start()
            result = _func(*args, **kwargs)
            _profiler.stop()
            if print_to_console:
                print(_profiler.output_text(unicode=True, color=True))
            if log_to_file:
                name = file_name if file_name is not None else _func.__name__ + ".profile"
                with open(f"{name}.html", "w") as f:
                    f.write(_profiler.output_html())
            return result
        return _wrapper

    if func:
        return wrapper(func)

    return wrapper


def timed(func):
    """
    Function decorator which logs execution times of the wrapped/annotated function at INFO level
    :param func: the function whose execution time to log
    :return: the wrapped function
    """
    @functools.wraps(func)
    def _wrapper(*args, **kwargs):
        start = int(round(time() * 1000))
        try:
            return func(*args, **kwargs)
        finally:
            end_ = int(round(time() * 1000)) - start
            log.info(f"Execution time of {func.__qualname__}: {end_ if end_ > 0 else 0} ms")
    return _wrapper


from abc import ABC, abstractmethod
from concurrent.futures.process import ProcessPoolExecutor
from typing import Sequence, List
from inspect import currentframe, getframeinfo

import pandas as pd

from ..vector_model import VectorModel
from ..util.multiprocessing import VectorModelWithSeparateFeatureGeneration
from ..util.pickle import PickleFailureDebugger


class EnsembleVectorModel(VectorModel, ABC):
    def __init__(self, models: Sequence[VectorModel], numProcesses=1):
        """
        :param models:
        :param numProcesses:
        """
        self.numProcesses = numProcesses
        self.models = list(models)
        super().__init__(checkInputColumns=False)

    def _fit(self, X: pd.DataFrame, Y: pd.DataFrame):
        if self.numProcesses == 1 or len(self.models) == 1:
            for model in self.models:
                model.fit(X, Y)
            return

        fittedModelFutures = []
        executor = ProcessPoolExecutor(max_workers=self.numProcesses)
        fitters = [VectorModelWithSeparateFeatureGeneration(model) for model in self.models]
        for fitter in fitters:
            intermediateStep = fitter.fitStart(X, Y)
            frameinfo = getframeinfo(currentframe())
            PickleFailureDebugger.logFailureIfEnabled(intermediateStep,
                contextInfo=f"Submitting {fitter} in {frameinfo.filename}:{frameinfo.lineno}")
            fittedModelFutures.append(executor.submit(intermediateStep.execute))
        for i, fittedModelFuture in enumerate(fittedModelFutures):
            self.models[i] = fitters[i].fitEnd(fittedModelFuture.result())

    def computeAllPredictions(self, X: pd.DataFrame):
        if self.numProcesses == 1 or len(self.models) == 1:
            return [model.predict(X) for model in self.models]

        predictionFutures = []
        executor = ProcessPoolExecutor(max_workers=self.numProcesses)
        predictors = [VectorModelWithSeparateFeatureGeneration(model) for model in self.models]
        for predictor in predictors:
            predictFinaliser = predictor.predictStart(X)
            frameinfo = getframeinfo(currentframe())
            PickleFailureDebugger.logFailureIfEnabled(predictFinaliser,
                contextInfo=f"Submitting {predictFinaliser} in {frameinfo.filename}:{frameinfo.lineno}")
            predictionFutures.append(executor.submit(predictFinaliser.execute))
        return [predictionFuture.result() for predictionFuture in predictionFutures]

    def _predict(self, x):
        predictionsDataFrames = self.computeAllPredictions(x)
        return self.aggregatePredictions(predictionsDataFrames)

    @abstractmethod
    def aggregatePredictions(self, predictionsDataFrames: List[pd.DataFrame]) -> pd.DataFrame:
        pass


class EnsembleRegressionVectorModel(EnsembleVectorModel, ABC):
    def isRegressionModel(self):
        return True


class EnsembleClassificationVectorModel(EnsembleVectorModel, ABC):
    def isRegressionModel(self):
        return False



from enum import Enum
from typing import Union

import numpy as np
import pandas as pd
import sklearn.preprocessing

from .util.dtype import toFloatArray


class NormalisationMode(Enum):
    NONE = "none"
    MAX_ALL = "max_all"
    MAX_BY_COLUMN = "max_by_column"
    STANDARDISED = "standardised"


class VectorDataScaler:
    def __init__(self, dataFrame: pd.DataFrame, normalisationMode: NormalisationMode):
        self.normalisationMode = normalisationMode
        self.scale, self.translate = self._computeScalingParams(dataFrame.values, normalisationMode)
        self.dimensionNames = list(dataFrame.columns)

    @classmethod
    def _computeScalingParams(cls, rawArray: np.ndarray, normalisationMode: NormalisationMode):
        """
        :param rawArray: numpy array containing raw data
        :param normalisationMode: the normalization mode (0=none, 1=by maximum in entire data set, 2=by separate maximum in each column)
        """
        translate = None
        scale = None
        if normalisationMode != NormalisationMode.NONE:
            if len(rawArray.shape) != 2:
                raise ValueError(f"Only 2D arrays are supported by {cls.__name__} with mode {normalisationMode}")
            dim = rawArray.shape[1]
            if normalisationMode == NormalisationMode.MAX_ALL:
                scale = np.ones(dim) * np.max(np.abs(rawArray))
            elif normalisationMode == NormalisationMode.MAX_BY_COLUMN:
                scale = np.ones(dim)
                for i in range(dim):
                    scale[i] = np.max(np.abs(rawArray[:, i]))
            elif normalisationMode == NormalisationMode.STANDARDISED:
                standardScaler = sklearn.preprocessing.StandardScaler()
                standardScaler.fit(rawArray)
                translate = standardScaler.mean_
                scale = standardScaler.scale_
            else:
                raise Exception("Unknown normalization mode")
        return scale, translate

    @staticmethod
    def _array(data: Union[pd.DataFrame, np.ndarray]):
        return toFloatArray(data)

    def getNormalisedArray(self, data: Union[pd.DataFrame, np.ndarray]) -> np.ndarray:
        result = self._array(data)
        if self.translate is not None:
            result = result - self.translate
        if self.scale is not None:
            result = result / self.scale
        return result

    def getDenormalisedArray(self, data: Union[pd.DataFrame, np.ndarray]) -> np.ndarray:
        result = self._array(data)
        if self.scale is not None:
            result = result * self.scale
        if self.translate is not None:
            result = result + self.translate
        return result



import enum
import logging
import pickle
import pandas as pd

from .cache import PersistentKeyValueCache, DelayedUpdateHook

log = logging.getLogger(__name__)


class MySQLPersistentKeyValueCache(PersistentKeyValueCache):

    class ValueType(enum.Enum):
        DOUBLE = ("DOUBLE", False)  # (SQL data type, isCachedValuePickled)
        BLOB = ("BLOB", True)

    def __init__(self, host, db, user, pw, valueType: ValueType, tableName="cache", deferredCommitDelaySecs=1.0, inMemory=False):
        import MySQLdb
        self.conn = MySQLdb.connect(host=host, database=db, user=user, password=pw)
        self.tableName = tableName
        self.maxKeyLength = 255
        self._updateHook = DelayedUpdateHook(self._commit, deferredCommitDelaySecs)
        self._numEntriesToBeCommitted = 0

        cacheValueSqlType, self.isCacheValuePickled = valueType.value

        cursor = self.conn.cursor()
        cursor.execute(f"SHOW TABLES;")
        if tableName not in [r[0] for r in cursor.fetchall()]:
            cursor.execute(f"CREATE TABLE {tableName} (cache_key VARCHAR({self.maxKeyLength}) PRIMARY KEY, cache_value {cacheValueSqlType});")
        cursor.close()

        self._inMemoryDf = None if not inMemory else self._loadTableToDataFrame()

    def _loadTableToDataFrame(self):
        df = pd.read_sql(f"SELECT * FROM {self.tableName};", con=self.conn, index_col="cache_key")
        if self.isCacheValuePickled:
            df["cache_value"] = df["cache_value"].apply(pickle.loads)
        return df

    def set(self, key, value):
        key = str(key)
        if len(key) > self.maxKeyLength:
            raise ValueError(f"Key too long, maximal key length is {self.maxKeyLength}")
        cursor = self.conn.cursor()
        cursor.execute(f"SELECT COUNT(*) FROM {self.tableName} WHERE cache_key=%s", (key, ))
        storedValue = pickle.dumps(value) if self.isCacheValuePickled else value
        if cursor.fetchone()[0] == 0:
            cursor.execute(f"INSERT INTO {self.tableName} (cache_key, cache_value) VALUES (%s, %s)",
                (key, storedValue))
        else:
            cursor.execute(f"UPDATE {self.tableName} SET cache_value=%s WHERE cache_key=%s", (storedValue, key))
        self._numEntriesToBeCommitted += 1
        self._updateHook.handleUpdate()
        cursor.close()
        if self._inMemoryDf is not None:
            self._inMemoryDf["cache_value"][str(key)] = value

    def get(self, key):
        value = self._getFromInMemoryDf(key)
        if value is None:
            value = self._getFromTable(key)
        return value

    def _getFromTable(self, key):
        cursor = self.conn.cursor()
        cursor.execute(f"SELECT cache_value FROM {self.tableName} WHERE cache_key=%s", (str(key), ))
        row = cursor.fetchone()
        if row is None:
            return None
        storedValue = row[0]
        value = pickle.loads(storedValue) if self.isCacheValuePickled else storedValue
        return value

    def _getFromInMemoryDf(self, key):
        if self._inMemoryDf is None:
            return None
        try:
            return self._inMemoryDf["cache_value"][str(key)]
        except Exception as e:
            log.debug(f"Unable to load value for key {str(key)} from in-memory dataframe: {e}")
            return None

    def _commit(self):
        log.info(f"Committing {self._numEntriesToBeCommitted} cache entries to the database")
        self.conn.commit()
        self._numEntriesToBeCommitted = 0



"""
This module contains various helper functions.
"""
from typing import Any, Sequence, Union, TypeVar, List


T = TypeVar("T")


def countNone(*args: Any) -> int:
    """
    Counts the number of arguments that are None

    :param args: various arguments
    :return: the number of arguments that are None
    """
    c = 0
    for a in args:
        if a is None:
            c += 1
    return c


def anyNone(*args: Any) -> bool:
    """
    :param args: various arguments
    :return: True if any of the arguments are None, False otherwise
    """
    return countNone(*args) > 0


def allNone(*args: Any) -> bool:
    """
    :param args: various arguments
    :return: True if all of the arguments are None, False otherwise
    """
    return countNone(*args) == len(args)


def markUsed(*args):
    """
    Utility function to mark identifiers as used.
    The function does nothing.

    :param args: pass identifiers that shall be marked as used here
    """
    pass


def flattenArguments(args: Sequence[Union[T, Sequence[T]]]) -> List[T]:
    """
    Main use case is to support both interfaces of the type f(T1, T2, ...) and f([T1, T2, ...]) simultaneously.
    It is assumed that if the latter form is passed, the arguments are either in a list or a tuple. Moreover,
    T cannot be a tuple or a list itself.

    Overall this function is not all too safe and one should be aware of what one is doing when using it
    """
    result = []
    for arg in args:
        if isinstance(arg, list) or isinstance(arg, tuple):
            result.extend(arg)
        else:
            result.append(arg)
    return result


from typing import Union, List, Dict, Any, Sequence, Iterable, Optional
import re


reCommaWhitespacePotentiallyBreaks = re.compile(r",\s+")


def dictString(d: Dict, brackets: Optional[str] = None):
    s = ', '.join([f'{k}={toString(v)}' for k, v in d.items()])
    if brackets is not None:
        return brackets[:1] + s + brackets[-1:]
    else:
        return s


def listString(l: Iterable[Any], brackets="[]"):
    return brackets[:1] + ", ".join((toString(x) for x in l)) + brackets[-1:]


def toString(x):
    if type(x) == list:
        return listString(x)
    elif type(x) == tuple:
        return listString(x, brackets="()")
    elif type(x) == dict:
        return dictString(x, brackets="{}")
    else:
        s = str(x)
        s = reCommaWhitespacePotentiallyBreaks.sub(", ", s)  # remove any unwanted line breaks and indentation after commas (as generated, for example, by sklearn objects)
        return s


def objectRepr(obj, memberNamesOrDict: Union[List[str], Dict[str, Any]]):
    if type(memberNamesOrDict) == dict:
        membersDict = memberNamesOrDict
    else:
        membersDict = {m: toString(getattr(obj, m)) for m in memberNamesOrDict}
    return f"{obj.__class__.__name__}[{dictString(membersDict)}]"


def orRegexGroup(allowedNames: Sequence[str]):
    """

    :param allowedNames: strings to include as literals in the regex
    :return: raw string of the type (<name1>| ...|<nameN>), where special characters in the names have been escaped
    """
    allowedNames = [re.escape(name) for name in allowedNames]
    return r"(%s)" % "|".join(allowedNames)


class ToStringMixin:
    """
    Provides default implementations for __str__ and __repr__ which contain all attribute names and their values. The
    latter also contains the object id.
    """

    def _toStringClassName(self):
        return type(self).__qualname__

    def _toStringProperties(self, exclude: Optional[Union[str, Iterable[str]]] = None, include: Optional[Union[str, Iterable[str]]] = None,
            excludeExceptions: Optional[List[str]] = None, includeForced: Optional[List[str]] = None,
            additionalEntries: Dict[str, Any] = None) -> str:
        """
        Creates a string of the class attributes, with optional exclusions/inclusions/additions.
        Exclusions take precedence over inclusions.

        :param exclude: attributes to be excluded
        :param include: attributes to be included; if None/empty, include all that are not excluded
        :param additionalEntries: additional key-value-pairs which are added to the string just like the other attributes
        :return: a string containing attribute names and values
        """
        def mklist(x):
            if x is None:
                return []
            if type(x) == str:
                return [x]
            return x

        exclude = mklist(exclude)
        include = mklist(include)
        includeForced = mklist(includeForced)
        excludeExceptions = mklist(excludeExceptions)

        def isExcluded(k):
            if k in includeForced or k in excludeExceptions:
                return False
            if k in exclude:
                return True
            if self._toStringExcludePrivate():
                isPrivate = k.startswith("_")
                return isPrivate
            else:
                return False

        # determine relevant attribute dictionary
        if len(include) == 0:  # exclude semantics (include everything by default)
            attributeDict = self.__dict__
        else:  # include semantics (include only inclusions)
            attributeDict = {k: getattr(self, k) for k in set(include + includeForced) if hasattr(self, k)}

        # apply exclusions and remove underscores from attribute names
        d = {k.strip("_"): v for k, v in attributeDict.items() if not isExcluded(k)}

        if additionalEntries is not None:
            d.update(additionalEntries)

        return dictString(d)

    def _toStringObjectInfo(self) -> str:
        """
        Creates a string containing information on the object instance which is to appear between the square brackets in the string
        representation, i.e. if the class name is Foo, then it is the asterisk in "Foo[*]".
        By default will make use of all the exclusions/inclusions that are specified by other member functions.
        This method can be overwritten by sub-classes to provide a custom string.

        :return: a string containing the desired content
        """
        return self._toStringProperties(exclude=self._toStringExcludes(), include=self._toStringIncludes(),
            excludeExceptions=self._toStringExcludeExceptions(), includeForced=self._toStringIncludesForced(),
            additionalEntries=self._toStringAdditionalEntries())

    def _toStringExcludes(self) -> List[str]:
        """
        Makes the string representation exclude the returned attributes.
        Returns a list of attribute names to be excluded from __str__ and __repr__. This method can be overwritten by
        sub-classes which can call super and extend the list returned.
        This method will only have no effect if _toStringObjectInfo is overridden to not use its result.

        :return: a list of attribute names
        """
        return []

    def _toStringIncludes(self) -> List[str]:
        """
        Makes the string representation include only the returned attributes (i.e. introduces inclusion semantics) - except
        if the list is empty, in which case all attributes are included by default.
        To add an included attribute in a sub-class, regardless of any super-classes using exclusion or inclusion semantics,
        use _toStringIncludesForced instead.
        This method can be overwritten by sub-classes which can call super and extend the list.
        This method will only have no effect if _toStringObjectInfo is overridden to not use its result.

        :return: a list of attribute names; if empty, include all attributes (except the ones being excluded according to other methods)
        """
        return []

    def _toStringIncludesForced(self) -> List[str]:
        """
        Defines a list of attribute names that are required to be present in the string representation, regardless of the
        instance using include semantics or exclude semantics, thus facilitating added inclusions in sub-classes.

        :return: a list of attribute names
        """
        return []

    def _toStringAdditionalEntries(self) -> Dict[str, Any]:
        return {}

    def _toStringExcludePrivate(self) -> bool:
        """
        :return: whether to exclude properties that are private, i.e. start with an underscore; explicitly included attributes
            will still be considered
        """
        return False

    def _toStringExcludeExceptions(self) -> List[str]:
        """
        Defines attribute names which should not be excluded even though other rules (e.g. the exclusion of private members
        via _toStringExcludePrivate) would otherwise exclude them.

        :return: a list of attribute names
        """
        return []

    def __str__(self):
        return f"{self._toStringClassName()}[{self._toStringObjectInfo()}]"

    def __repr__(self):
        info = f"id={id(self)}"
        propertyInfo = self._toStringObjectInfo()
        if len(propertyInfo) > 0:
            info += ", " + propertyInfo
        return f"{self._toStringClassName()}[{info}]"


def prettyStringRepr(s: Any, initialIndentationLevel=0, indentationString="    "):
    """
    Creates a pretty string representation (using indentations) from the given object/string representation (as generated, for example, via
    ToStringMixin). An indentation level is added for every opening bracket.

    :param s: an object or object string representation
    :param initialIndentationLevel: the initial indentation level
    :param indentationString: the string which corresponds to a single indentation level
    :return: a reformatted version of the input string with added indentations and line break
    """
    if type(s) != str:
        s = str(s)
    indent = initialIndentationLevel
    result = indentationString * indent
    i = 0

    def nl():
        nonlocal result
        result += "\n" + (indentationString * indent)

    def take(cnt=1):
        nonlocal result, i
        result += s[i:i+cnt]
        i += cnt

    def findMatching(j):
        start = j
        op = s[j]
        cl = {"[": "]", "(": ")", "'": "'"}[s[j]]
        isBracket = cl != s[j]
        stack = 0
        while j < len(s):
            if s[j] == op and (isBracket or j == start):
                stack += 1
            elif s[j] == cl:
                stack -= 1
            if stack == 0:
                return j
            j += 1
        return None

    brackets = "[("
    quotes = "'"
    while i < len(s):
        isBracket = s[i] in brackets
        isQuote = s[i] in quotes
        if isBracket or isQuote:
            iMatch = findMatching(i)
            takeFullMatchWithoutBreak = False
            if iMatch is not None:
                k = iMatch + 1
                takeFullMatchWithoutBreak = not isBracket or (k-i <= 60 and not("=" in s[i:k] and "," in s[i:k]))
                if takeFullMatchWithoutBreak:
                    take(k-i)
            if not takeFullMatchWithoutBreak:
                take(1)
                indent += 1
                nl()
        elif s[i] in "])":
            take(1)
            indent -= 1
        elif s[i:i+2] == ", ":
            take(2)
            nl()
        else:
            take(1)

    return result


import logging
from abc import ABC, abstractmethod
from typing import Tuple, Sequence, Optional, Union, List, Iterator

import numpy as np
import pandas as pd
import sklearn.preprocessing
import torch
from torch.autograd import Variable

from .. import normalisation
from ..data import DataFrameSplitter, DataFrameSplitterFractional
from ..data_transformation import DFTSkLearnTransformer
from ..util.dtype import toFloatArray
from ..util.pickle import setstate


log = logging.getLogger(__name__)


def toTensor(d: Union[torch.Tensor, np.ndarray, list], cuda=False):
    if not isinstance(d, torch.Tensor):
        if isinstance(d, np.ndarray):
            d = torch.from_numpy(d)
        elif isinstance(d, list):
            d = torch.from_numpy(np.array(d))
        else:
            raise ValueError()
    if cuda:
        d.cuda()
    return d


class TensorScaler(ABC):
    @abstractmethod
    def cuda(self):
        """
        Makes this scaler's components use CUDA
        """
        pass

    @abstractmethod
    def normalise(self, tensor: torch.Tensor) -> torch.Tensor:
        """
        Applies scaling/normalisation to the given tensor
        :param tensor: the tensor to scale/normalise
        :return: the scaled/normalised tensor
        """
        pass

    @abstractmethod
    def denormalise(self, tensor: torch.Tensor) -> torch.Tensor:
        """
        Applies the inverse of method normalise to the given tensor
        :param tensor: the tensor to denormalise
        :return: the denormalised tensor
        """
        pass


class TensorScalerCentreAndScale(TensorScaler):
    def __init__(self, centre: Optional[torch.Tensor] = None, scale: Optional[torch.Tensor] = None):
        self.centre = centre
        self.scale = scale

    def cuda(self):
        if self.scale is not None:
            self.scale = self.scale.cuda()
        if self.centre is not None:
            self.centre = self.centre.cuda()

    def normalise(self, tensor: torch.Tensor) -> torch.Tensor:
        if self.centre is not None:
            tensor -= self.centre
        if self.scale is not None:
            tensor *= self.scale
        return tensor

    def denormalise(self, tensor: torch.Tensor) -> torch.Tensor:
        if self.scale is not None:
            tensor /= self.scale
        if self.centre is not None:
            tensor += self.centre
        return tensor


class TensorScalerFromVectorDataScaler(TensorScalerCentreAndScale):
    def __init__(self, vectorDataScaler: normalisation.VectorDataScaler, cuda: bool):
        if vectorDataScaler.scale is not None:
            invScale = torch.from_numpy(vectorDataScaler.scale).float()
            scale = 1.0 / invScale
        else:
            scale = None
        centre = vectorDataScaler.translate
        if centre is not None:
            centre = torch.from_numpy(vectorDataScaler.translate).float()
        super().__init__(centre=centre, scale=scale)
        if cuda:
            self.cuda()

    def __setstate__(self, state):
        if "translate" in state:
            if state["scale"] is not None:  # old representation where scale is actually inverse scale
                state["scale"] = 1.0 / state["scale"]
        setstate(TensorScalerFromVectorDataScaler, self, state, renamedProperties={"translate": "centre"})


class TensorScalerIdentity(TensorScaler):
    def cuda(self):
        pass

    def normalise(self, tensor: torch.Tensor) -> torch.Tensor:
        return tensor

    def denormalise(self, tensor: torch.Tensor) -> torch.Tensor:
        return tensor


class TensorScalerFromDFTSkLearnTransformer(TensorScalerCentreAndScale):
    def __init__(self, dft: DFTSkLearnTransformer):
        trans = dft.sklearnTransformer
        if isinstance(trans, sklearn.preprocessing.RobustScaler):
            centre = trans.center_
            scale = trans.scale_
            isReciprocalScale = True
        else:
            raise ValueError(f"sklearn transformer of type '{trans.__class__}' is unhandled")
        if centre is not None:
            centre = torch.from_numpy(centre).float()
        if scale is not None:
            scale = torch.from_numpy(scale).float()
            if isReciprocalScale:
                scale = 1.0 / scale
        super().__init__(centre=centre, scale=scale)


class Tensoriser(ABC):
    """
    Represents a method for transforming a data frame into one or more tensors to be processed by a neural network model
    """
    def tensorise(self, df: pd.DataFrame) -> Union[torch.Tensor, List[torch.Tensor]]:
        result = self._tensorise(df)
        if type(result) == list:
            lengths = set(map(len, result))
            if len(lengths) != 1:
                raise Exception("Lengths of tensors inconsistent")
            length = lengths.pop()
        else:
            length = len(result)
        if length != len(df):
            raise Exception(f"{self} produced result of length {length} for DataFrame of shape {df.shape}")
        return result

    @abstractmethod
    def _tensorise(self, df: pd.DataFrame) -> Union[torch.Tensor, List[torch.Tensor]]:
        pass

    @abstractmethod
    def fit(self, df: pd.DataFrame, model=None):
        """
        :param df: the data frame with which to fit this tensoriser
        :param model: the model in the context of which the fitting takes place (if any).
            The fitting process may set parameters within the model that can only be determined from the (pre-tensorised) data.
        """
        pass


class RuleBasedTensoriser(Tensoriser, ABC):
    """
    Base class for tensorisers which transform data frames into tensors based on a predefined set of rules and do not require fitting
    """
    def fit(self, df: pd.DataFrame, model=None):
        pass


class TensoriserDataFrameFloatValuesMatrix(RuleBasedTensoriser):
    def _tensorise(self, df: pd.DataFrame) -> np.ndarray:
        return torch.from_numpy(toFloatArray(df)).float()


class TensoriserClassLabelIndices(RuleBasedTensoriser):
    def _tensorise(self, df: pd.DataFrame) -> np.ndarray:
        if len(df.columns) != 1:
            raise ValueError("Expected a single column containing the class label indices")
        return torch.from_numpy(df[df.columns[0]].values).long()


class DataUtil(ABC):
    """Interface for DataUtil classes, which are used to process data for neural networks"""

    @abstractmethod
    def splitIntoTensors(self, fractionalSizeOfFirstSet) -> Tuple[Tuple[torch.Tensor, torch.Tensor], Tuple[torch.Tensor, torch.Tensor]]:
        """
        Splits the data set

        :param fractionalSizeOfFirstSet: the desired fractional size in
        :return: a tuple (A, B) where A and B are tuples (in, out) with input and output data
        """
        pass

    @abstractmethod
    def getOutputTensorScaler(self) -> TensorScaler:
        """
        Gets the scaler with which to scale model outputs

        :return: the scaler
        """
        pass

    @abstractmethod
    def getInputTensorScaler(self) -> TensorScaler:
        """
        Gets the scaler with which to scale model inputs

        :return: the scaler
        """
        pass

    @abstractmethod
    def modelOutputDim(self) -> int:
        """
        :return: the dimensionality that is to be output by the model to be trained
        """
        pass

    @abstractmethod
    def inputDim(self):
        """
        :return: the dimensionality of the input or None if it is variable
        """
        pass


class VectorDataUtil(DataUtil):
    def __init__(self, inputs: pd.DataFrame, outputs: pd.DataFrame, cuda: bool, normalisationMode=normalisation.NormalisationMode.NONE,
            differingOutputNormalisationMode=None, inputTensoriser: Optional[Tensoriser] = None, outputTensoriser: Optional[Tensoriser] = None,
            dataFrameSplitter: Optional[DataFrameSplitter] = None):
        """
        :param inputs: the data frame of inputs
        :param outputs: the data frame of outputs
        :param cuda: whether to apply CUDA
        :param normalisationMode: the normalisation mode to use for inputs and (unless differingOutputNormalisationMode is specified) outputs
        :param differingOutputNormalisationMode: the normalisation mode to apply to outputs, overriding normalisationMode;
            if None, use normalisationMode
        """
        if inputs.shape[0] != outputs.shape[0]:
            raise ValueError("Output length must be equal to input length")
        self.inputs = inputs
        self.outputs = outputs
        self.inputTensoriser = inputTensoriser if inputTensoriser is not None else TensoriserDataFrameFloatValuesMatrix()
        self.outputTensoriser = outputTensoriser if outputTensoriser is not None else TensoriserDataFrameFloatValuesMatrix()
        self.inputVectorDataScaler = normalisation.VectorDataScaler(self.inputs, normalisationMode)
        self.inputTensorScaler = TensorScalerFromVectorDataScaler(self.inputVectorDataScaler, cuda)
        self.outputVectorDataScaler = normalisation.VectorDataScaler(self.outputs, normalisationMode if differingOutputNormalisationMode is None else differingOutputNormalisationMode)
        self.outputTensorScaler = TensorScalerFromVectorDataScaler(self.outputVectorDataScaler, cuda)
        self.dataFrameSplitter = dataFrameSplitter

    def __len__(self):
        return len(self.inputs)

    def getOutputTensorScaler(self):
        return self.outputTensorScaler

    def getInputTensorScaler(self):
        return self.inputTensorScaler

    def _computeSplitIndices(self, fractionalSizeOfFirstSet):
        splitter = self.dataFrameSplitter
        if splitter is None:
            # By default, we use a simple fractional split without shuffling.
            # Shuffling is usually unnecessary, because in evaluation contexts, the data may have already been shuffled by the evaluator
            # (unless explicitly disabled by the user). Furthermore, not shuffling gives the user the possibility to manually
            # order the data in ways that result in desirable fractional splits (though the user may, of course, simply override
            # the splitter to achieve any desired split).
            splitter = DataFrameSplitterFractional(shuffle=False)
        indices_A, indices_B = splitter.computeSplitIndices(self.inputs, fractionalSizeOfFirstSet)
        return indices_A, indices_B

    def splitIntoTensors(self, fractionalSizeOfFirstSet):
        indices_A, indices_B = self._computeSplitIndices(fractionalSizeOfFirstSet)
        A = self._tensorsForIndices(indices_A)
        B = self._tensorsForIndices(indices_B)
        return A, B

    def _dataFramesForIndices(self, indices):
        inputDF = self.inputs.iloc[indices]
        outputDF = self.outputs.iloc[indices]
        return inputDF, outputDF

    def _tensorsForIndices(self, indices):
        inputDF, outputDF = self._dataFramesForIndices(indices)
        return self._tensorsForDataFrames(inputDF, outputDF)

    def _tensorsForDataFrames(self, inputDF, outputDF):
        # apply normalisation (if any)
        if self.inputVectorDataScaler.normalisationMode != normalisation.NormalisationMode.NONE:
            inputDF = pd.DataFrame(self.inputVectorDataScaler.getNormalisedArray(inputDF), columns=inputDF.columns, index=inputDF.index)
        if self.outputVectorDataScaler.normalisationMode != normalisation.NormalisationMode.NONE:
            outputDF = pd.DataFrame(self.outputVectorDataScaler.getNormalisedArray(outputDF), columns=outputDF.columns, index=outputDF.index)

        return self.inputTensoriser.tensorise(inputDF), self.outputTensoriser.tensorise(outputDF)

    def splitIntoDataSets(self, fractionalSizeOfFirstSet, cuda: bool, tensoriseDynamically=False) -> Tuple["TorchDataSet", "TorchDataSet"]:
        if not tensoriseDynamically:
            (xA, yA), (xB, yB) = self.splitIntoTensors(fractionalSizeOfFirstSet)
            return TorchDataSetFromTensors(xA, yA, cuda), TorchDataSetFromTensors(xB, yB, cuda)
        else:
            if self.inputVectorDataScaler.normalisationMode != normalisation.NormalisationMode.NONE or \
                    self.outputVectorDataScaler.normalisationMode != normalisation.NormalisationMode.NONE:
                raise Exception("Dynamic tensorisation is not supported when using data scaling")
            indicesA, indicesB = self._computeSplitIndices(fractionalSizeOfFirstSet)
            inputA, outputA = self._dataFramesForIndices(indicesA)
            inputB, outputB = self._dataFramesForIndices(indicesB)
            dsA = TorchDataSetFromDataFramesDynamicallyTensorised(inputA, outputA, cuda, inputTensoriser=self.inputTensoriser,
                outputTensoriser=self.outputTensoriser)
            dsB = TorchDataSetFromDataFramesDynamicallyTensorised(inputB, outputB, cuda, inputTensoriser=self.inputTensoriser,
                outputTensoriser=self.outputTensoriser)
            return dsA, dsB

    def inputDim(self):
        return self.inputs.shape[1]

    def outputDim(self):
        """
        :return: the dimensionality of the outputs (ground truth values)
        """
        return self.outputs.shape[1]

    def modelOutputDim(self):
        return self.outputDim()


class ClassificationVectorDataUtil(VectorDataUtil):
    def __init__(self, inputs: pd.DataFrame, outputs: pd.DataFrame, cuda, numClasses, normalisationMode=normalisation.NormalisationMode.NONE,
            inputTensoriser: Tensoriser = None, outputTensoriser: Tensoriser = None, dataFrameSplitter: Optional[DataFrameSplitter] = None):
        if len(outputs.columns) != 1:
            raise Exception(f"Exactly one output dimension (the class index) is required, got {len(outputs.columns)}")
        super().__init__(inputs, outputs, cuda, normalisationMode=normalisationMode,
            differingOutputNormalisationMode=normalisation.NormalisationMode.NONE, inputTensoriser=inputTensoriser,
            outputTensoriser=TensoriserClassLabelIndices() if outputTensoriser is None else outputTensoriser,
            dataFrameSplitter=dataFrameSplitter)
        self.numClasses = numClasses

    def modelOutputDim(self):
        return self.numClasses


class TorchDataSet:
    @abstractmethod
    def iterBatches(self, batchSize: int, shuffle: bool = False, inputOnly=False) -> Iterator[Union[Tuple[torch.Tensor, torch.Tensor],
            Tuple[Sequence[torch.Tensor], torch.Tensor], torch.Tensor, Sequence[torch.Tensor]]]:
        """
        Provides an iterator over batches from the data set.

        :param batchSize: the maximum size of each batch
        :param shuffle: whether to shuffle the data set
        :param inputOnly: whether to provide only inputs (rather than inputs and corresponding outputs).
            If true, provide only inputs, where inputs can either be a tensor or a tuple of tensors.
            If false, provide a pair (i, o) with inputs and corresponding outputs (o is always a tensor).
            Some data sets may only be able to provide inputs, in which case inputOnly=False should lead to an
            exception.
        """
        pass

    @abstractmethod
    def size(self) -> Optional[int]:
        """
        Returns the total size of the data set (number of data points) if it is known.

        :return: the number of data points or None of the size is not known.
        """
        pass


class TorchDataSetProvider:
    def __init__(self, inputTensorScaler: Optional[TensorScaler] = None, outputTensorScaler: Optional[TensorScaler] = None,
            inputDim: Optional[int] = None, modelOutputDim: int = None):
        if inputTensorScaler is None:
            inputTensorScaler = TensorScalerIdentity()
        if outputTensorScaler is None:
            outputTensorScaler = TensorScalerIdentity()
        if modelOutputDim is None:
            raise ValueError("The model output dimension must be provided")
        self.inputTensorScaler = inputTensorScaler
        self.outputTensorScaler = outputTensorScaler
        self.inputDim = inputDim
        self.modelOutputDim = modelOutputDim

    @abstractmethod
    def provideSplit(self, fractionalSizeOfFirstSet: float) -> Tuple[TorchDataSet, TorchDataSet]:
        """
        Provides two data sets, which could, for example, serve as training and validation sets.

        :param fractionalSizeOfFirstSet: the fractional size of the first data set
        :return: a tuple of data sets (A, B) where A has (approximately) the given fractional size and B encompasses
            the remainder of the data
        """
        pass

    def getOutputTensorScaler(self) -> TensorScaler:
        return self.outputTensorScaler

    def getInputTensorScaler(self) -> TensorScaler:
        return self.inputTensorScaler

    def getModelOutputDim(self) -> int:
        """
        :return: the number of output dimensions that would be required to be generated by the model to match this dataset.
        """
        return self.modelOutputDim

    def getInputDim(self) -> Optional[int]:
        """
        :return: the number of output dimensions that would be required to be generated by the model to match this dataset.
            For models that accept variable input sizes (such as RNNs), this may be None.
        """
        return self.inputDim


class TensorTuple:
    """
    Represents a tuple of tensors (or a single tensor) and can be used to manipulate the contained tensors simultaneously
    """
    def __init__(self, tensors: Union[torch.Tensor, Sequence[torch.Tensor]]):
        if isinstance(tensors, torch.Tensor):
            tensors = [tensors]
        lengths = set(map(len, tensors))
        if len(lengths) != 1:
            raise ValueError("Not all tensors are of the same length")
        self.length = lengths.pop()
        self.tensors = tensors

    def __len__(self):
        return self.length

    def __getitem__(self, key) -> "TensorTuple":
        t = tuple((t[key] for t in self.tensors))
        return TensorTuple(t)

    def cuda(self) -> "TensorTuple":
        return TensorTuple([t.cuda() for t in self.tensors])

    def tuple(self) -> Sequence[torch.Tensor]:
        return tuple(self.tensors)

    def item(self) -> Union[torch.Tensor, Sequence[torch.Tensor]]:
        if len(self.tensors) == 1:
            return self.tensors[0]
        else:
            return self.tuple()

    def concat(self, other: "TensorTuple") -> "TensorTuple":
        if len(self.tensors) != len(other.tensors):
            raise ValueError("Tensor tuples are incompatible")
        tensors = [torch.cat([a, b], dim=0) for a, b in zip(self.tensors, other.tensors)]
        return TensorTuple(tensors)


class TorchDataSetFromTensors(TorchDataSet):
    def __init__(self, x: Union[torch.Tensor, Sequence[torch.Tensor]], y: Optional[torch.Tensor], cuda: bool):
        """
        :param x: the input tensor(s); if more than one, they must be of the same length (and a slice of each shall be provided to the
            model as an input in each batch)
        :param y: the output tensor
        :param cuda: whether any generated tensors shall be moved to the selected CUDA device
        """
        x = TensorTuple(x)
        y = TensorTuple(y) if y is not None else None
        if y is not None and len(x) != len(y):
            raise ValueError("Tensors are not of the same length")
        self.x = x
        self.y = y
        self.cuda = cuda

    def iterBatches(self, batchSize: int, shuffle: bool = False, inputOnly=False) -> Iterator[Union[Tuple[torch.Tensor, torch.Tensor],
            Tuple[Sequence[torch.Tensor], torch.Tensor], torch.Tensor, Sequence[torch.Tensor]]]:
        tensorTuples = (self.x, self.y) if not inputOnly and self.y is not None else (self.x,)
        yield from self._get_batches(tensorTuples, batchSize, shuffle)

    def _get_batches(self, tensorTuples: Sequence[TensorTuple], batch_size, shuffle):
        length = len(tensorTuples[0])
        if shuffle:
            index = torch.randperm(length)
        else:
            index = torch.LongTensor(range(length))
        start_idx = 0
        while start_idx < length:
            end_idx = min(length, start_idx + batch_size)
            excerpt = index[start_idx:end_idx]
            batch = []
            for tensorTuple in tensorTuples:
                if len(tensorTuple) != length:
                    raise Exception("Passed tensors of differing lengths")
                t = tensorTuple[excerpt]
                if self.cuda:
                    t = t.cuda()
                item = t.item()
                if type(item) == tuple:
                    item = tuple(Variable(t) for t in item)
                else:
                    item = Variable(item)
                batch.append(item)
            if len(batch) == 1:
                yield batch[0]
            else:
                yield tuple(batch)
            start_idx += batch_size

    def size(self):
        return len(self.x)


class TorchDataSetFromDataFramesPreTensorised(TorchDataSetFromTensors):
    def __init__(self, input: pd.DataFrame, output: Optional[pd.DataFrame], cuda: bool,
            inputTensoriser: Optional[Tensoriser] = None, outputTensoriser: Optional[Tensoriser] = None):
        if inputTensoriser is None:
            inputTensoriser = TensoriserDataFrameFloatValuesMatrix()
        log.debug(f"Applying {inputTensoriser} to data frame of length {len(input)} ...")
        inputTensors = inputTensoriser.tensorise(input)
        if output is not None:
            if outputTensoriser is None:
                outputTensoriser = TensoriserDataFrameFloatValuesMatrix()
            log.debug(f"Applying {outputTensoriser} to data frame of length {len(output)} ...")
            outputTensors = outputTensoriser.tensorise(output)
        else:
            outputTensors = None
        super().__init__(inputTensors, outputTensors, cuda)


class TorchDataSetFromDataFramesDynamicallyTensorised(TorchDataSet):
    def __init__(self, input: pd.DataFrame, output: Optional[pd.DataFrame], cuda: bool,
            inputTensoriser: Optional[Tensoriser] = None, outputTensoriser: Optional[Tensoriser] = None):
        self.inputDF = input
        self.outputDF = output
        self.cuda = cuda
        if inputTensoriser is None:
            inputTensoriser = TensoriserDataFrameFloatValuesMatrix()
        self.inputTensoriser = inputTensoriser
        if output is not None:
            if len(input) != len(output):
                raise ValueError("Lengths of input and output data frames must be equal")
            if outputTensoriser is None:
                outputTensoriser = TensoriserDataFrameFloatValuesMatrix()
        self.outputTensoriser = outputTensoriser

    def size(self) -> Optional[int]:
        return len(self.inputDF)

    def iterBatches(self, batchSize: int, shuffle: bool = False, inputOnly=False):
        length = len(self.inputDF)
        if shuffle:
            index = torch.randperm(length)
        else:
            index = torch.LongTensor(range(length))
        i = 0
        while i < length:
            batchIndices = index[i:i+batchSize]
            inputTensors = TensorTuple(self.inputTensoriser.tensorise(self.inputDF.iloc[batchIndices]))
            if self.cuda:
                inputTensors = inputTensors.cuda()
            if inputOnly:
                yield inputTensors.item()
            else:
                outputTensors = TensorTuple(self.outputTensoriser.tensorise(self.outputDF.iloc[batchIndices]))
                if self.cuda:
                    outputTensors = outputTensors.cuda()
                yield inputTensors.item(), outputTensors.item()
            i += batchSize


class TorchDataSetFromDataFrames(TorchDataSet):
    def __init__(self, input: pd.DataFrame, output: Optional[pd.DataFrame], cuda: bool,
            inputTensoriser: Optional[Tensoriser] = None, outputTensoriser: Optional[Tensoriser] = None,
            tensoriseDynamically=False):
        if tensoriseDynamically:
            self._torchDataSet: TorchDataSet = TorchDataSetFromDataFramesDynamicallyTensorised(input, output, cuda,
                inputTensoriser=inputTensoriser, outputTensoriser=outputTensoriser)
        else:
            self._torchDataSet: TorchDataSet = TorchDataSetFromDataFramesPreTensorised(input, output, cuda,
                inputTensoriser=inputTensoriser, outputTensoriser=outputTensoriser)

    def iterBatches(self, batchSize: int, shuffle: bool = False, inputOnly=False):
        yield from self._torchDataSet.iterBatches(batchSize, shuffle=shuffle, inputOnly=inputOnly)

    def size(self) -> Optional[int]:
        return self._torchDataSet.size()


class TorchDataSetProviderFromDataUtil(TorchDataSetProvider):
    def __init__(self, dataUtil: DataUtil, cuda: bool):
        super().__init__(inputTensorScaler=dataUtil.getInputTensorScaler(), outputTensorScaler=dataUtil.getOutputTensorScaler(),
            inputDim=dataUtil.inputDim(), modelOutputDim=dataUtil.modelOutputDim())
        self.dataUtil = dataUtil
        self.cuda = cuda

    def provideSplit(self, fractionalSizeOfFirstSet: float) -> Tuple[TorchDataSet, TorchDataSet]:
        (x1, y1), (x2, y2) = self.dataUtil.splitIntoTensors(fractionalSizeOfFirstSet)
        return TorchDataSetFromTensors(x1, y1, self.cuda), TorchDataSetFromTensors(x2, y2, self.cuda)


class TensorTransformer(ABC):
    @abstractmethod
    def transform(self, t: torch.Tensor) -> torch.Tensor:
        pass


import logging
from typing import Callable, Optional, Sequence, Union

import torch.nn.functional

from .mlp_modules import MultiLayerPerceptron
from ...torch_base import VectorTorchModel, TorchVectorRegressionModel, TorchVectorClassificationModel, ClassificationOutputMode
from ...torch_enums import ActivationFunction
from ...torch_opt import NNOptimiserParams
from .... import NormalisationMode

log: logging.Logger = logging.getLogger(__name__)


class MultiLayerPerceptronTorchModel(VectorTorchModel):
    def __init__(self, cuda: bool, hiddenDims: Sequence[int], hidActivationFunction: Callable[[torch.Tensor], torch.Tensor],
            outputActivationFunction: Optional[Callable[[torch.Tensor], torch.Tensor]], pDropout: Optional[float] = None) -> None:
        super().__init__(cuda=cuda)
        self.hidActivationFunction = ActivationFunction.torchFunctionFromAny(hidActivationFunction)
        self.outputActivationFunction = ActivationFunction.torchFunctionFromAny(outputActivationFunction)
        self.hiddenDims = hiddenDims
        self.pDropout = pDropout

    def createTorchModuleForDims(self, inputDim: int, outputDim: int) -> torch.nn.Module:
        return MultiLayerPerceptron(inputDim, outputDim, self.hiddenDims,
            hidActivationFn=self.hidActivationFunction, outputActivationFn=self.outputActivationFunction,
            pDropout=self.pDropout)


class MultiLayerPerceptronVectorRegressionModel(TorchVectorRegressionModel):
    def __init__(self, hiddenDims: Sequence[int] = (5, 5), hidActivationFunction: Callable[[torch.Tensor], torch.Tensor] = torch.sigmoid,
            outputActivationFunction: Optional[Callable[[torch.Tensor], torch.Tensor]] = None,
            normalisationMode: NormalisationMode = NormalisationMode.MAX_BY_COLUMN,
            cuda: bool = True, pDropout: Optional[float] = None, nnOptimiserParams: Optional[NNOptimiserParams] = None,
            **nnOptimiserDictParams) -> None:
        """
        :param hiddenDims: sequence containing the number of neurons to use in hidden layers
        :param hidActivationFunction: the activation function (torch.nn.functional.* or torch.*) to use for all hidden layers
        :param outputActivationFunction: the output activation function (torch.nn.functional.* or torch.* or None)
        :param normalisationMode: the normalisation mode to apply to input and output data
        :param cuda: whether to use CUDA (GPU acceleration)
        :param pDropout: the probability with which to apply dropouts after each hidden layer
        :param nnOptimiserParams: parameters for NNOptimiser; if None, use default (or what is specified in nnOptimiserDictParams)
        :param nnOptimiserDictParams: [for backward compatibility] parameters for NNOptimiser (alternative to nnOptimiserParams)
        """
        nnOptimiserParams = NNOptimiserParams.fromEitherDictOrInstance(nnOptimiserDictParams, nnOptimiserParams)
        super().__init__(MultiLayerPerceptronTorchModel, [cuda, hiddenDims, hidActivationFunction, outputActivationFunction],
                dict(pDropout=pDropout), normalisationMode, nnOptimiserParams)


class MultiLayerPerceptronVectorClassificationModel(TorchVectorClassificationModel):
    def __init__(self, hiddenDims: Sequence[int] = (5, 5),
            hidActivationFunction: Callable[[torch.Tensor], torch.Tensor] = torch.sigmoid,
            outputActivationFunction: Optional[Union[Callable[[torch.Tensor], torch.Tensor], str, ActivationFunction]] = torch.nn.functional.log_softmax,
            normalisationMode: NormalisationMode = NormalisationMode.MAX_BY_COLUMN, cuda: bool = True, pDropout: Optional[float] = None,
            nnOptimiserParams: Optional[NNOptimiserParams] = None, **nnOptimiserDictParams) -> None:
        """
        :param hiddenDims: sequence containing the number of neurons to use in hidden layers
        :param hidActivationFunction: the activation function (torch.nn.functional.* or torch.*) to use for all hidden layers
        :param outputActivationFunction: the output activation function (function from torch.nn.functional.*, function name, enum instance or None)
        :param normalisationMode: the normalisation mode to apply to input and output data
        :param cuda: whether to use CUDA (GPU acceleration)
        :param pDropout: the probability with which to apply dropouts after each hidden layer
        :param nnOptimiserParams: parameters for NNOptimiser; if None, use default (or what is specified in nnOptimiserDictParams)
        :param nnOptimiserDictParams: [for backward compatibility] parameters for NNOptimiser (alternative to nnOptimiserParams)
        """
        nnOptimiserParams = NNOptimiserParams.fromEitherDictOrInstance(nnOptimiserDictParams, nnOptimiserParams)
        outputMode = ClassificationOutputMode.forActivationFn(ActivationFunction.torchFunctionFromAny(outputActivationFunction))
        super().__init__(outputMode, MultiLayerPerceptronTorchModel, [cuda, hiddenDims, hidActivationFunction, outputActivationFunction],
            dict(pDropout=pDropout), normalisationMode, nnOptimiserParams)
